from .command import CmdTypes, MrgModes
from .result import Res,ResTypes
from .diff import DifferenceFinder
from ..query.query import Qry,Obj,ObjSeg
from ..query.queryrunner import QryRunner
from ..util.error import EoqError
from ..action import CmdRunnerBasedCallManager
from ..event.event import EvtProvider,ChgEvt,ChgTypes,ALL_EVENT_TYPES
from ..util.util import IsListOfObjects,IsNoList,IsList,ApplyToAllElements
from ..util.logger import NoLogging
from ..util.benchmark import Benchmark
from ..serialization.jsonserializer import JsonSerializer
from ..serialization.textserializer import TextSerializer
from ..validation.validation import ValidationManager

from datetime import time
from uuid import uuid4

import traceback


class Session:
    def __init__(self,sessionId,sessionNumber):
        self.sessionId = sessionId
        self.sessionNumber = sessionNumber
        self.user = None
        self.begin = time
        self.isTemp = False
        #initialize event table
        self.events = {} #lists the events this session is observing as 
        for e in ALL_EVENT_TYPES:
            self.events[e] = {} #indicates to listen to no key at all
            
class Transaction:
    def __init__(self,tid):
        self.tid = tid
        self.session = None
        self.changes = [] #dictionary which collects changes until the transaction is finished
        self.history = []
        self.isMuted = False
        
'''
 Cmd Runner
'''   
class CmdRunner(EvtProvider):
    def __init__(self,mdbAccessor,maxChanges=10000,logger=NoLogging(),enableBenchmark=False):
        super().__init__()
        self.mdbAccessor = mdbAccessor
        self.maxChanges = maxChanges #the number of changes that are preserved
        #logging
        self.logger = logger
        self.enableBenchmark = enableBenchmark
        self.logSerializer = JsonSerializer()
        #validation
        self.validationManager = ValidationManager(self)
        self.textSerializer = TextSerializer()
        #initialize internals
        self.latestTransactionId = 0
        self.sessionCount = 0
        self.transactions = {}
        #sessions
        self.sessions = {} #dict containing sessions keys and related session informations
        #changes
        self.earliestChangeId = 0
        self.latestChangeId = 0
        self.changes = {}
        #query runner
        self.qryEvaluator = QryRunner(self.mdbAccessor,self.validationManager,enableBenchmark=enableBenchmark)
        self.callManager = CmdRunnerBasedCallManager(self)        
        #init command evaluators functor table
        self.cmdEvaluators = {}
        self.cmdEvaluators[CmdTypes.GET] = self.ExecGet
        self.cmdEvaluators[CmdTypes.SET] = self.ExecSet
        self.cmdEvaluators[CmdTypes.ADD] = self.ExecAdd
        self.cmdEvaluators[CmdTypes.REM] = self.ExecRem
        self.cmdEvaluators[CmdTypes.MOV] = self.ExecMov
        self.cmdEvaluators[CmdTypes.CLO] = self.ExecClo
        self.cmdEvaluators[CmdTypes.CRT] = self.ExecCrt
        self.cmdEvaluators[CmdTypes.CRN] = self.ExecCrn
        self.cmdEvaluators[CmdTypes.QRF] = self.ExecQrf
        self.cmdEvaluators[CmdTypes.CPR] = self.ExecCpr
        self.cmdEvaluators[CmdTypes.MRG] = self.ExecMrg
        self.cmdEvaluators[CmdTypes.STS] = self.ExecSts
        self.cmdEvaluators[CmdTypes.GMM] = self.ExecGmm
        self.cmdEvaluators[CmdTypes.RMM] = self.ExecRmm
        self.cmdEvaluators[CmdTypes.UMM] = self.ExecUmm
        self.cmdEvaluators[CmdTypes.GAA] = self.ExecGaa
        self.cmdEvaluators[CmdTypes.HEL] = self.ExecHel
        self.cmdEvaluators[CmdTypes.SES] = self.ExecSes
        self.cmdEvaluators[CmdTypes.GBY] = self.ExecGby
        self.cmdEvaluators[CmdTypes.CHG] = self.ExecChg
        self.cmdEvaluators[CmdTypes.OBS] = self.ExecObs
        self.cmdEvaluators[CmdTypes.UBS] = self.ExecUbs
        self.cmdEvaluators[CmdTypes.CAL] = self.ExecCal
        self.cmdEvaluators[CmdTypes.ASC] = self.ExecAsc
        self.cmdEvaluators[CmdTypes.ABC] = self.ExecAbc
        self.cmdEvaluators[CmdTypes.ADC] = self.ExecAdc
        self.cmdEvaluators[CmdTypes.RMC] = self.ExecRmc
        self.cmdEvaluators[CmdTypes.VAL] = self.ExecVal
        self.cmdEvaluators[CmdTypes.GAC] = self.ExecGac
        self.cmdEvaluators[CmdTypes.CMP] = self.ExecCmp
        self.cmdEvaluators[CmdTypes.MUT] = self.ExecMut
        self.cmdEvaluators[CmdTypes.UMT] = self.ExecUmt
        
        #start listening to external events
        if(self.mdbAccessor):
            self.mdbAccessor.Observe(self.OnCallMdbAccessorEvent)
        if(self.callManager):
            self.callManager.Observe(self.OnCallManagerEvent)
            
        #benchmark
        if(self.enableBenchmark):
            self.benchmark = Benchmark()
            allCmdTypes = [getattr(CmdTypes,k) for k in CmdTypes.__dict__ if not k.startswith('_')]
            for t in allCmdTypes:
                self.benchmark.InitMessure(t)


        # merge / diff
        self.oldRoot = None
        self.newRoot = None
        self.postponedMergeReferences = []
        self.diffcommands = []
        self.mDummies = 1
        
    '''
    COMMAND EXECUTION
    '''
    
    def Exec(self,cmd,sessionId=None):
        res = None
        try:
            tid = self.StartTransaction(sessionId)
            res = self.ExecOnTransaction(cmd,tid)
        except Exception as e: 
            errorMsg = "Command %s failed: %s"%(self.logSerializer.Ser(cmd),str(e))
            self.logger.Error(errorMsg)
            traceback.print_exc()
            res = Res(cmd.cmd,ResTypes.ERR,str(e),tid,self.latestChangeId)     
        self.EndTransaction(res,tid)
        return res
    
    def ExecOnTransaction(self,cmd,tid):
        self.logger.PassivatableLog('transaction',lambda : self.logSerializer.Ser(cmd))
        cmdType = cmd.cmd
        try:
            evaluator = self.cmdEvaluators[cmdType]
        except KeyError:
            raise EoqError(0,"Error evaluating command: Unknown command type: %s."%(cmd.cmd))
        if self.enableBenchmark: self.benchmark.Start()
        try:
            res = evaluator(cmd.a,tid)
        except Exception as e: 
            if self.enableBenchmark: self.benchmark.Stop(cmdType)
            raise e #forward exception
        if self.enableBenchmark: self.benchmark.Stop(cmdType)
        self.logger.PassivatableLog('transaction',lambda: self.logSerializer.Ser(res))
        return res
    
    def ExecCmp(self,args,tid):
        transaction = self.GetTransaction(tid)
        status = ResTypes.OKY
        subresults = []
        n = 0;
        for cmd in args:
            n = n+1;
            try:
                subresult = self.ExecOnTransaction(cmd, tid) 
                if(transaction.isMuted):
                    subresult.v = None #cut of the output, but it is still in the history
                subresults.append(subresult)
            except Exception as e:
                raise EoqError(0,"Sub command %d failed: %s"%(n,str(e)))
        return Res(CmdTypes.CMP,status,subresults,tid,self.latestChangeId)
    
    def ExecMut(self,args,tid):
        transaction = self.GetTransaction(tid)
        transaction.isMuted = True
        res = True
        self.AddToHistory(res,tid)
        return Res(CmdTypes.MUT,ResTypes.OKY,res,tid,self.latestChangeId)
    
    def ExecUmt(self,args,tid):
        transaction = self.GetTransaction(tid)
        transaction.isMuted = False
        res = True
        self.AddToHistory(res,tid)
        return Res(CmdTypes.MUT,ResTypes.OKY,res,tid,self.latestChangeId)
            
    def ExecGet(self,args,tid):
        status = ResTypes.OKY
        target = args
        history = self.GetHistory(tid)
        res = self.qryEvaluator.Eval(target,history)
        self.AddToHistory(res,tid)
        #result = []
        #if not(isinstance(res, list)):
            #res = [res]
        #for constraint in res:
            #try:
                #result.append([constraint.constraintType, constraint.target, constraint.feature, constraint.law, constraint.name, constraint.annotation])
            #except:
                #result.append(constraint)
        if isinstance(res,list):
            res = self.ChangeConstrainttoList(res)
        return Res(CmdTypes.GET,status,res,tid,self.latestChangeId)
    
    def ExecSet(self,args,tid):
        status = ResTypes.OKY
        #res = None  #[target,feature,oldVal]
        cid = self.latestChangeId
        #eval all arguments
        history = self.GetHistory(tid)
        target = self.qryEvaluator.Eval(args[0],history)
        feature = self.qryEvaluator.Eval(args[1],history)
        value = self.qryEvaluator.Eval(args[2],history)
        
        #set the value(s) depending on the multiplicity of the arguments
        if(IsNoList(target)): # e.g. #20
            if(IsNoList(feature)):
                (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Set(target,feature,value)
                cid = self.AddLocalChange(tid,ChgTypes.SET,target,feature,value,oldVal,oldOwner,oldFeature,oldIndex)
                self.Val(target, tid)
            elif(IsListOfObjects(feature)):
                if(IsNoList(value)):
                    for f in feature:
                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Set(target,f,value)
                        cid = self.AddLocalChange(tid,ChgTypes.SET,target,f,value,oldVal,oldOwner,oldFeature,oldIndex)
                        self.Val(target, tid)
                elif(IsListOfObjects(value) and len(value) == len(feature)):
                    for i in range(len(feature)):
                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Set(target,feature[i],value[i])
                        cid = self.AddLocalChange(tid,ChgTypes.SET,target,feature[i],value[i],oldVal,oldOwner,oldFeature,oldIndex)
                        self.Val(target, tid)
                else:
                    raise EoqError(0,'Error in set: value must be single value or list of values with equal length of the number of features, but got: %s:'%(value)) 
            else:
                raise EoqError(0,'Error in set: feature must be single object or list of objects but got: %s:'%(feature)) 
        elif(IsListOfObjects(target)): # e.g. [#20,#22,#23]
            if(IsNoList(feature)):
                if(IsNoList(value)):
                    for t in target:
                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Set(t,feature,value)
                        cid = self.AddLocalChange(tid,ChgTypes.SET,t,feature,value,oldVal,oldOwner,oldFeature,oldIndex)
                        self.Val(t, tid)
                elif(IsListOfObjects(value) and len(value) == len(target)):
                    for i in range(len(target)):
                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Set(target[i],feature,value[i])
                        cid = self.AddLocalChange(tid,ChgTypes.SET,target[i],feature,value[i],oldVal,oldOwner,oldFeature,oldIndex)
                        self.Val(target[i], tid)
                else:
                    raise EoqError(0,'Error in set: value must be single value or list of values with equal length of the number of targets, but got: %s:'%(value)) 
            elif(IsListOfObjects(feature)):
                if(IsNoList(value)):
                    for t in target:
                        for f in feature:
                            (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Set(t,f,value)
                            cid = self.AddLocalChange(tid,ChgTypes.SET,t,f,value,oldVal,oldOwner,oldFeature,oldIndex)
                            self.Val(t, tid)
                elif(IsListOfObjects(value) and len(value) == len(feature)):
                    for t in target:
                        for i in range(len(feature)):
                            (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Set(t,feature[i],value[i])
                            cid = self.AddLocalChange(tid,ChgTypes.SET,t,feature[i],value[i],oldVal,oldOwner,oldFeature,oldIndex)
                            self.Val(t, tid)
                elif(IsList(value) and len(value) == len(target)):
                    for j in range(len(target)):
                        if(IsListOfObjects(value[j]) and len(value[j]) == len(feature)):
                            for i in range(len(feature)):
                                (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Set(target[j],feature[i],value[j][i])
                                cid = self.AddLocalChange(tid,ChgTypes.SET,target[j],feature[i],value[j][i],oldVal,oldOwner,oldFeature,oldIndex)
                                self.Val(target[j], tid)
                        else:
                            raise EoqError(0,'Error in set: for multiple targets and multiple features the value for each entry must have the same length as the number of features. Expected %d entries for target %d, but got %d.'%(len(feature),j,len(value[j])))
                else:
                    raise EoqError(0,'Error in set: value must be single value or list of values with equal length of the number of targets, but got: %s:'%(value)) 
            else:
                raise EoqError(0,'Error in set: feature must be single object or list of objects but got: %s:'%(feature)) 
        else: 
            raise EoqError(0,'Error in set: target must be single object or list of objects but got: %s:'%(target))
        
        res = [target,feature,value]
        self.AddToHistory(res,tid)
        return Res(CmdTypes.SET,status,res,tid,cid)
    
    def ExecAdd(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        cid = self.latestChangeId
        #eval all arguments
        history = self.GetHistory(tid)
        target = self.qryEvaluator.Eval(args[0],history)
        feature = self.qryEvaluator.Eval(args[1],history)
        value = self.qryEvaluator.Eval(args[2],history)
        # reset the validation cache since the model is changed
        self.validationManager.ClearCache()
        
        #set the value(s) depending on the multiplicity of the arguments
        if(IsNoList(target)): # e.g. #20
            if(IsNoList(feature)):
                if(IsNoList(value)):
                    (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Add(target,feature,value)
                    cid = self.AddLocalChange(tid,ChgTypes.ADD,target,feature,value,oldVal,oldOwner,oldFeature,oldIndex)
                    self.Val(target, tid)
                elif(IsList(value)):
                    for v in value:
                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Add(target,feature,v)
                        cid = self.AddLocalChange(tid,ChgTypes.ADD,target,feature,v,oldVal,oldOwner,oldFeature,oldIndex)
                        self.Val(target, tid)
                else:
                    raise EoqError(0,'Error in add: value must be single value or list of values, but got: %s:'%(value)) 
            elif(IsListOfObjects(feature)):
                if(IsNoList(value)):
                    for f in feature:
                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Add(target,f,value)
                        cid = self.AddLocalChange(tid,ChgTypes.ADD,target,f,value,oldVal,oldOwner,oldFeature,oldIndex)
                        self.Val(target, tid)
                elif(IsList(value) and len(value) == len(feature)):
                    for i in range(len(feature)):
                        if(IsListOfObjects(value[i])):
                            for v in value[i]:
                                (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Add(target,feature[i],v)
                                cid = self.AddLocalChange(tid,ChgTypes.ADD,target,feature[i],v,oldVal,oldOwner,oldFeature,oldIndex)
                                self.Val(target, tid)
                        else:
                            raise EoqError(0,'Error in add: for multiple features the value must be a list of list of objects for each feature, but entry %d is %s.'%(i,value[i]))
                else:
                    raise EoqError(0,'Error in add: value must be single value or list of list of values with outer list having a length equal to the number of features, but got: %s:'%(value)) 
            else:
                raise EoqError(0,'Error in add: feature must be single object or list of objects but got: %s:'%(feature)) 
        elif(IsListOfObjects(target)): # e.g. [#20,#22,#23]
            if(IsNoList(feature)):
                if(IsNoList(value)):
                    for t in target:
                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Add(t,feature,value)
                        cid = self.AddLocalChange(tid,ChgTypes.ADD,t,feature,value,oldVal,oldOwner,oldFeature,oldIndex)
                        self.Val(t, tid)
                elif(IsList(value) and len(value) == len(target)):
                    for j in range(len(target)):
                        if(IsListOfObjects(value[j])):
                            for i in range(len(value[j])):
                                (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Add(target[j],feature,value[j][i])
                                cid = self.AddLocalChange(tid,ChgTypes.ADD,target[j],feature,value[j][i],oldVal,oldOwner,oldFeature,oldIndex)
                                self.Val(target[j], tid)
                        else:
                            raise EoqError(0,'Error in add: for multiple targets the value must be a list of list of objects for each target, but entry %d is %s.'%(j,value[j]))
                else:
                    raise EoqError(0,'Error in add: value must be single value or list of list of values with the outer list having a length equal to the number of targets, but got: %s:'%(value)) 
            elif(IsListOfObjects(feature)):
                if(IsNoList(value)):
                    for t in target:
                        for f in feature:
                            (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Add(t,f,value)
                            cid = self.AddLocalChange(tid,ChgTypes.ADD,t,f,value,oldVal,oldOwner,oldFeature,oldIndex)
                            self.Val(t, tid)
                elif(IsListOfObjects(value) and len(value) == len(feature)):
                    for t in target:
                        for i in range(len(feature)):
                            (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Add(t,feature[i],value[i])
                            cid = self.AddLocalChange(tid,ChgTypes.ADD,t,feature[i],value[i],oldVal,oldOwner,oldFeature,oldIndex)
                            self.Val(t, tid)
                elif(IsList(value) and len(value) == len(target)):
                    for j in range(len(target)):
                        if(IsListOfObjects(value[j]) and len(value[j]) == len(feature)):
                            for i in range(len(feature)):
                                (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Add(target[j],feature[i],value[j][i])
                                cid = self.AddLocalChange(tid,ChgTypes.ADD,target[j],feature[i],value[j][i],oldVal,oldOwner,oldFeature,oldIndex)
                                self.Val(target[j], tid)
                        elif(IsList(value[j]) and len(value[j]) == len(feature)):
                            for i in range(len(feature)):
                                if(IsList(value[j][i])):
                                    for v in value[j][i]:
                                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Add(target[j],feature[i],v)
                                        cid = self.AddLocalChange(tid,ChgTypes.ADD,target[j],feature[i],v,oldVal,oldOwner,oldFeature,oldIndex)
                                        self.Val(target[j], tid)
                                else:
                                    raise EoqError(0,'Error in add: for multiple targets, multiple features and multiple values value must list equal to targets containing a list equal to features containing a list of values, but got %s for target %d and feature %d.'%(value[j][i],j,i)) 
                        else:
                            raise EoqError(0,'Error in add: for multiple targets and multiple features the value for each entry must have the same length as the number of features. Expected %d entries for target %d, but got %s.'%(len(feature),j,value[j]))
                else:
                    raise EoqError(0,'Error in add: value must be single value or list of list of list of values with outer list having a length equal to the number of targets and the middle list with a length equal to the number of features, but got: %s:'%(value)) 
            else:
                raise EoqError(0,'Error in add: feature must be single object or list of objects but got: %s:'%(feature)) 
        else: 
            raise EoqError(0,'Error in add: target must be single object or list of objects but got: %s:'%(target))
        
        res = [target,feature,value]
        self.AddToHistory(res,tid)
        return Res(CmdTypes.ADD,status,res,tid,cid)
    
    def ExecRem(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        cid = self.latestChangeId
        #eval all arguments
        history = self.GetHistory(tid)
        target = self.qryEvaluator.Eval(args[0],history)
        feature = self.qryEvaluator.Eval(args[1],history)
        value = self.qryEvaluator.Eval(args[2],history)
        # reset the validation cache since the model is changed
        self.validationManager.ClearCache()
        
        #set the value(s) depending on the multiplicity of the arguments
        if(IsNoList(target)): # e.g. #20
            if(IsNoList(feature)):
                if(IsNoList(value)):
                    (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Remove(target,feature,value)
                    cid = self.AddLocalChange(tid,ChgTypes.REM,target,feature,value,oldVal,oldOwner,oldFeature,oldIndex)
                    self.Val(target, tid)
                elif(IsList(value)):
                    for v in value:
                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Remove(target,feature,v)
                        cid = self.AddLocalChange(tid,ChgTypes.REM,target,feature,v,oldVal,oldOwner,oldFeature,oldIndex)
                        self.Val(target, tid)
                else:
                    raise EoqError(0,'Error in remove: value must be single value or list of values, but got: %s:'%(value)) 
            elif(IsListOfObjects(feature)):
                if(IsNoList(value)):
                    for f in feature:
                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Remove(target,f,value)
                        cid = self.AddLocalChange(tid,ChgTypes.REM,target,f,value,oldVal,oldOwner,oldFeature,oldIndex)
                        self.Val(target, tid)
                elif(IsList(value) and len(value) == len(feature)):
                    for i in range(len(feature)):
                        if(IsListOfObjects(value[i])):
                            for v in value[i]:
                                (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Remove(target,feature[i],v)
                                cid = self.AddLocalChange(tid,ChgTypes.REM,target,feature[i],v,oldVal,oldOwner,oldFeature,oldIndex)
                                self.Val(target, tid)
                        else:
                            raise EoqError(0,'Error in remove: for multiple features the value must be a list of list of objects for each feature, but entry %d is %s.'%(i,value[i]))
                else:
                    raise EoqError(0,'Error in remove: value must be single value or list of list of values with outer list having a length equal to the number of features, but got: %s:'%(value)) 
            else:
                raise EoqError(0,'Error in remove: feature must be single object or list of objects but got: %s:'%(feature)) 
        elif(IsListOfObjects(target)): # e.g. [#20,#22,#23]
            if(IsNoList(feature)):
                if(IsNoList(value)):
                    for t in target:
                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Remove(t,feature,value)
                        cid = self.AddLocalChange(tid,ChgTypes.REM,t,feature,value,oldVal,oldOwner,oldFeature,oldIndex)
                        self.Val(t, tid)
                elif(IsList(value) and len(value) == len(target)):
                    for j in range(len(target)):
                        if(IsListOfObjects(value[j])):
                            for i in range(len(value[j])):
                                (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Remove(target[j],feature,value[j][i])
                                cid = self.AddLocalChange(tid,ChgTypes.REM,target[j],feature,value[j][i],oldVal,oldOwner,oldFeature,oldIndex)
                                self.Val(target[j], tid)
                        else:
                            raise EoqError(0,'Error in add: for multiple targets the value must be a list of list of objects for each target, but entry %d is %s.'%(j,value[j]))
                else:
                    raise EoqError(0,'Error in remove: value must be single value or list of list of values with the outer list having a length equal to the number of targets, but got: %s:'%(value)) 
            elif(IsListOfObjects(feature)):
                if(IsNoList(value)):
                    for t in target:
                        for f in feature:
                            (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Remove(t,f,value)
                            cid = self.AddLocalChange(tid,ChgTypes.REM,t,f,value,oldVal,oldOwner,oldFeature,oldIndex)
                            self.Val(t, tid)
                elif(IsListOfObjects(value) and len(value) == len(feature)):
                    for t in target:
                        for i in range(len(feature)):
                            (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Remove(t,feature[i],value[i])
                            cid = self.AddLocalChange(tid,ChgTypes.REM,t,feature[i],value[i],oldVal,oldOwner,oldFeature,oldIndex)
                            self.Val(t, tid)
                elif(IsList(value) and len(value) == len(target)):
                    for j in range(len(target)):
                        if(IsListOfObjects(value[j]) and len(value[j]) == len(feature)):
                            for i in range(len(feature)):
                                (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Remove(target[j],feature[i],value[j][i])
                                cid = self.AddLocalChange(tid,ChgTypes.REM,target[j],feature[i],value[j][i],oldVal,oldOwner,oldFeature,oldIndex)
                                self.Val(target[j], tid)
                        elif(IsList(value[j]) and len(value[j]) == len(feature)):
                            for i in range(len(feature)):
                                if(IsList(value[j][i])):
                                    for v in value[j][i]:
                                        (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Remove(target[j],feature[i],v)
                                        cid = self.AddLocalChange(tid,ChgTypes.REM,target[j],feature[i],v,oldVal,oldOwner,oldFeature,oldIndex)
                                        self.Val(target[j], tid)
                                else:
                                    raise EoqError(0,'Error in remove: for multiple targets, multiple features and multiple values value must list equal to targets containing a list equal to features containing a list of values, but got %s for target %d and feature %d.'%(value[j][i],j,i))
                        else:
                            raise EoqError(0,'Error in remove: for multiple targets and multiple features the value for each entry must have the same length as the number of features. Expected %d entries for target %d, but got %d.'%(len(feature),j,value[j]))
                else:
                    raise EoqError(0,'Error in remove: value must be single value or list of list of list of values with outer list having a lenght equal to the number of targets and the middle list with a length equal to the number of features, but got: %s:'%(value)) 
            else:
                raise EoqError(0,'Error in remove: feature must be single object or list of objects but got: %s:'%(feature)) 
        else: 
            raise EoqError(0,'Error in remove: target must be single object or list of objects but got: %s:'%(target))
        
        res = [target,feature,value]
        self.AddToHistory(res,tid)
        return Res(CmdTypes.REM,status,res,tid,cid)
    
    
    def ExecMov(self,args,tid):
        status = ResTypes.OKY
        #res = None  #[target,feature,oldVal]
        cid = self.latestChangeId
        #eval all arguments
        history = self.GetHistory(tid)
        target = self.qryEvaluator.Eval(args[0],history)
        newIndex = self.qryEvaluator.Eval(args[1],history)
        
        #set the value(s) depending on the multiplicity of the arguments
        if(IsNoList(target)): # e.g. #20
            if(IsNoList(newIndex)):
                (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Move(target,newIndex)
                cid = self.AddLocalChange(tid,ChgTypes.MOV,target,oldFeature,newIndex,oldVal,oldOwner,oldFeature,oldIndex)
                self.Val(target, tid)
            else:
                raise EoqError(0,'Error in move: new index must be an integer, but got: %s:'%(newIndex)) 
        elif(IsListOfObjects(target)): # e.g. [#20,#22,#23]
            if(IsNoList(newIndex)):
                for t in target:
                    (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Move(t,newIndex)
                    cid = self.AddLocalChange(tid,ChgTypes.MOV,t,oldFeature,newIndex,oldVal,oldOwner,oldFeature,oldIndex)
                    self.Val(t, tid)
            elif(IsListOfObjects(newIndex) and len(newIndex) == len(target)):
                for i in range(len(target)):
                    (oldVal,oldOwner,oldFeature,oldIndex) = self.mdbAccessor.Move(target[i],newIndex[i])
                    cid = self.AddLocalChange(tid,ChgTypes.MOV,target[i],oldFeature,newIndex[i],oldVal,oldOwner,oldFeature,oldIndex)
                    self.Val(target[i], tid)
            else:
                raise EoqError(0,'Error in move: new index must be single value or list of values with equal length of the number of targets, but got: %s:'%(newIndex)) 
        else: 
            raise EoqError(0,'Error in move: target must be single object or list of objects but got: %s:'%(target))
        
        res = [target,newIndex]
        self.AddToHistory(res,tid)
        return Res(CmdTypes.MOV,status,res,tid,cid)
    
    
    def ExecClo(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        history = self.GetHistory(tid)
        target = self.qryEvaluator.Eval(args[0],history)
        mode = self.qryEvaluator.Eval(args[1],history)
        if(IsNoList(target)): # e.g. #20
            if(IsNoList(mode)):
                res = self.mdbAccessor.Clone(target,mode)
            else:
                raise EoqError(0,'Error in clone: mode must be a string, but got: %s:'%(mode)) 
        elif(IsListOfObjects(target)): # e.g. [#20,#22,#23]
            if(IsNoList(mode)):
                res = [self.mdbAccessor.Clone(t,mode) for t in target]
            elif(IsListOfObjects(mode) and len(mode) == len(target)):
                res = [self.mdbAccessor.Clone(target[i],mode[i]) for i in range(len(target))]
            else:
                raise EoqError(0,'Error in clone: mode must be a string or a list of strings, but got: %s:'%(mode)) 
        else:
            raise EoqError(0,'Error in clone: target must be an object or a list of objects, but got: %s:'%(target)) 
        self.AddToHistory(res,tid)
        return Res(CmdTypes.CLO,status,res,tid,self.latestChangeId)

    def ExecMrg(self, args, tid):
        '''
        Merges all changes from file 'new' onto file 'old'.
        '''
        status = ResTypes.OKY
        res = self.ExecCpr(args, tid)
        cmds = res.GetValue()
        # execute commands
        for cmd in cmds:
            self.ExecOnTransaction(cmd, tid)
        return Res(CmdTypes.MRG, status, res, tid, self.latestChangeId)

    def ExecCpr(self, args, tid):
        '''
        Creates a list with all the necessary commands to adapt 
        the old object to the new one.
        '''
        status = ResTypes.OKY
        res = None
        history = self.GetHistory(tid)
        old     = self.qryEvaluator.Eval(args[0], history) # old file
        new     = self.qryEvaluator.Eval(args[1], history) # new file
        mrgMode = self.qryEvaluator.Eval(args[2], history) # new file
        if IsList(old) or IsList(new):
            raise EoqError(0,'Error in Diff/Merge: No Lists allowed.')
        # init diff.py
        differenceFinder = DifferenceFinder(self, self.mdbAccessor, 
            old, new, tid, mrgMode, logger = self.logger)
        if mrgMode == MrgModes.DET:
            differenceFinder.CreateObjectOverviewFile()
        # start diff.py
        differenceFinder.Run()
        # obtain commands
        res = differenceFinder.GetCommands()
        # serialize commands
        if mrgMode in [MrgModes.LST, MrgModes.LOG, MrgModes.DET]:
            path = 'log/cmdList_{}_{}.txt'.format(old, new)
            with open(path, 'w') as file:
                file.write('### Commands to merge {} onto {} ###\n'.format(new, old))
                for cmd in res:
                    file.write(self.logSerializer.Ser(cmd))
                    file.write('\n')
        return Res(CmdTypes.CPR, status, res, tid, self.latestChangeId)
    
    def ExecCrt(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        history = self.GetHistory(tid)
        clazz = self.qryEvaluator.Eval(args[0],history)
        n = self.qryEvaluator.Eval(args[1],history)
        constructorArgs = self.qryEvaluator.Eval(args[2],history)
        #constructorArgs = [self.qryEvaluator.Eval(a,history) for a in args[2]]
        if(IsNoList(clazz)): # e.g. #20
            if(IsNoList(n)):
                res = self.mdbAccessor.Create(clazz,n,constructorArgs)
            else:
                raise EoqError(0,'Error in create: n must be a positive integer, but got: %s:'%(n)) 
        else:
            raise EoqError(0,'Error in create: clazz must be an object of type class, but got: %s:'%(clazz)) 
        self.AddToHistory(res,tid)
        return Res(CmdTypes.CRT,status,res,tid,self.latestChangeId)
    
    def ExecCrn(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        history = self.GetHistory(tid)
        packageName = self.qryEvaluator.Eval(args[0],history)
        className = self.qryEvaluator.Eval(args[1],history)
        n = self.qryEvaluator.Eval(args[2],history)
        constructorArgs = self.qryEvaluator.Eval(args[3],history)
        #constructorArgs = [self.qryEvaluator.Eval(a,history) for a in args[3]]
        if(IsNoList(packageName) and IsNoList(className)): # e.g. #20
            if(IsNoList(n)):
                res = self.mdbAccessor.CreateByName(packageName,className,n,constructorArgs)
            else:
                raise EoqError(0,'Error in create: n must be a positive integer, but got: %s:'%(n)) 
        else:
            raise EoqError(0,'Error in create: packageName and className must be strings, but got: %s, %s:'%(packageName,className)) 
        self.AddToHistory(res,tid)
        return Res(CmdTypes.CRN,status,res,tid,self.latestChangeId)
    
    def ExecQrf(self,args,tid):
        def qrfFunctor(a):
            res = None
            if(isinstance(a,ObjSeg)):
                #traverse the path backwards from the object to the root
                pathFromParent = []
                c = a
                p = self.mdbAccessor.GetParent(c)
                while(p):
                    f = self.mdbAccessor.GetContainingFeature(c)
                    n = self.mdbAccessor.Get(f,'name')
                    i = self.mdbAccessor.GetIndex(c)
                    pathFromParent.append((n,i))
                    #go one level back, i.e. the new parent is the parent of the old
                    c = p
                    p = self.mdbAccessor.GetParent(c)
                #build a query that leads to the object
                res = Qry()
                for segment in reversed(pathFromParent):
                    n = segment[0]
                    i = segment[1]
                    if(None==i):
                        res.Pth(n)
                    else:
                        res.Pth(n).Idx(i)
            else:
                res = a 
            return res
        
        status = ResTypes.OKY
        target = args
        history = self.GetHistory(tid)
        elements = self.qryEvaluator.Eval(target,history)
        
        res = ApplyToAllElements(elements, qrfFunctor)
        
        self.AddToHistory(res,tid)
        return Res(CmdTypes.GET,status,res,tid,self.latestChangeId)
    
    def ExecGmm(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        res = self.mdbAccessor.GetAllMetamodels()
        self.AddToHistory(res,tid)
        return Res(CmdTypes.GMM,status,res,tid,self.latestChangeId)
    
    def ExecRmm(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        history = self.GetHistory(tid)
        package = self.qryEvaluator.Eval(args,history) 
        res = self.mdbAccessor.RegisterMetamodel(package)
        self.AddToHistory(res,tid)
        return Res(CmdTypes.RMM,status,res,tid,self.latestChangeId)
    
    def ExecUmm(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        history = self.GetHistory(tid)
        package = self.qryEvaluator.Eval(args,history) 
        res = self.mdbAccessor.UnregisterMetamodel(package)
        self.AddToHistory(res,tid)
        return Res(CmdTypes.UMM,status,res,tid,self.latestChangeId)
    
    def ExecHel(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        transaction = self.GetTransaction(tid)
        history = transaction.history
        user = self.qryEvaluator.Eval(args[0],history) 
        password = self.qryEvaluator.Eval(args[1],history) 
        
        #prepare a new session
        sessionId = None
        #check identification
        if(1): #TODO: validate user and password
            session = None
            if(transaction.session):
                session = transaction.session
            else: #new session is needed
                session = self.InitSession(transaction, sessionId)
            sessionId = session.sessionId #retrieve the existing session id
            session.user = user
        else:
            raise EoqError(0,'Identification failed.')
        res = sessionId
        self.AddToHistory(res,tid)
        return Res(CmdTypes.HEL,status,res,tid,self.latestChangeId)
    
    def ExecSes(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        transaction = self.GetTransaction(tid)
        history = transaction.history
        sessionId = self.qryEvaluator.Eval(args,history) 
        try:
            session = self.sessions[sessionId]
            transaction.session = session
        except KeyError:
            raise EoqError(0,'Unknown session %s'%(sessionId))
        res = True
        self.AddToHistory(res,tid)
        return Res(CmdTypes.SES,status,res,tid,self.latestChangeId)
    
    def ExecGby(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        transaction = self.GetTransaction(tid)
        history = transaction.history
        sessionId = self.qryEvaluator.Eval(args,history) 
        try:
            self.sessions.pop(sessionId) #delete the session
            transaction.session = None #remove any eventual remaining reference to the session
        except KeyError:
            raise EoqError(0,'Unknown session %s'%(sessionId))
        res = True
        self.AddToHistory(res,tid)
        return Res(CmdTypes.SES,status,res,tid,self.latestChangeId)
    
    def ExecSts(self,args,tid):
        status = ResTypes.OKY
        #res = None   
        res = self.latestChangeId-1 #latest the latest change is one below the indication
        self.AddToHistory(res,tid)
        return Res(CmdTypes.STS,status,res,tid,self.latestChangeId)
    
    def ExecChg(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        history = self.GetHistory(tid)
        changeId = self.qryEvaluator.Eval(args[0],history)
        n = self.qryEvaluator.Eval(args[1],history)
        #sanity check
        if(n<0):
            raise EoqError(0,"Index must be greater or equal zero, but got: %d"%(n))
        
        if(changeId<self.earliestChangeId):
            if(n==0 or changeId+n >= self.latestChangeId): 
                res = [self.changes[i].a for i in range(self.earliestChangeId,self.latestChangeId)]
            elif(changeId+n >= self.earliestChangeId):
                res = [self.changes[i].a for i in range(self.earliestChangeId,changeId+n)]
            else: 
                res = []
        elif(changeId >= self.latestChangeId):
            res = []
        else: #is a valid change id
            if(n==0 or changeId+n >= self.latestChangeId):
                res = [self.changes[i].a for i in range(changeId,self.latestChangeId)]
            else:
                res = [self.changes[i].a for i in range(changeId,changeId+n)]    
        self.AddToHistory(res,tid)
        return Res(CmdTypes.CHG,status,res,tid,self.latestChangeId)
    
    def ExecObs(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        transaction = self.GetTransaction(tid)
        history = transaction.history
        eventType = str(self.qryEvaluator.Eval(args[0],history)) #everything must be string
        eventKey = str(self.qryEvaluator.Eval(args[1],history)) #everything must be string
        #check is session is known
        if(not transaction.session):
            raise EoqError(0,'Must specify session before observing events.')
        #add the desired events to the observation list
        sessionId = transaction.session.sessionId
        self.Obs(sessionId,eventType,eventKey)
        res = True
        self.AddToHistory(res,tid)
        return Res(CmdTypes.OBS,status,res,tid,self.latestChangeId)
    
    def ExecUbs(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        transaction = self.GetTransaction(tid)
        history = transaction.history
        eventType = str(self.qryEvaluator.Eval(args[0],history)) #everything must be string
        eventKey = str(self.qryEvaluator.Eval(args[1],history)) #everything must be string
        #check is session is known
        if(not transaction.session):
            raise EoqError(0,'Must specify session before observing events.')
        #remove the desired events to the observation list
        sessionId = transaction.session.sessionId
        self.Ubs(sessionId,eventType,eventKey)
        res = True
        self.AddToHistory(res,tid)
        return Res(CmdTypes.UBS,status,res,tid,self.latestChangeId)
    
    def ExecGaa(self,args,tid):
        status = ResTypes.OKY
        #res = None  
        #history = self.GetHistory(tid)
        res = [[a.name,[[p.name,p.type,p.min,p.max,p.default,p.description,[[o.value,o.description] for o in p.options]] for p in a.args],[[r.name,r.type,r.min,r.max,r.default,r.description,[[o.value,o.description] for o in r.options]] for r in a.results],a.description,a.tags] for a in self.callManager.GetAllActions()]
        self.AddToHistory(res,tid)
        return Res(CmdTypes.CAL,status,res,tid,self.latestChangeId)
    
    def ExecCal(self,args,tid):
        status = ResTypes.OKY  
        #evaluate any query within the arguments of the call
        transaction = self.GetTransaction(tid)
        history = transaction.history 
        name = self.qryEvaluator.Eval(args[0],history)
        actionArgs = []
        for qry in args[1]:
            aa = self.qryEvaluator.Eval(qry,history)
            actionArgs.append(aa)
            
        optsArr = args[2]
        
        session = transaction.session
        sessionId = session.sessionId if session else None
            
        #call the handler
        res = self.callManager.RunCallSync(name,actionArgs,optsArr,tid,sessionId)
        self.AddToHistory(res,tid)
        return Res(CmdTypes.CAL,status,res,tid,self.latestChangeId)
    
    def ExecAsc(self,args,tid):
        status = ResTypes.OKY
        transaction = self.GetTransaction(tid)
        history = transaction.history 
        
        name = self.qryEvaluator.Eval(args[0],history)
        actionArgs = []
        for qry in args[1]:
            aa = self.qryEvaluator.Eval(qry,history)
            actionArgs.append(aa)
            
        optsArr = args[2]
            
        # Create a temporary session for the action call. This is necessary,
        # to prevent the original session ID from being exposed to the call
        # and for enabling a self-controlled event listening within the action.
        inheritSession = transaction.session
        callSession = self.CreateNewSession(inheritSession=inheritSession,isTemp=True)  
        
        callSessionId =   callSession.sessionId
        inheritSessionId = inheritSession.sessionId if inheritSession else None
        
        res = self.callManager.RunCallAsync(name,actionArgs,optsArr,tid,callSessionId,inheritSessionId)
        self.AddToHistory(res,tid)
        return Res(CmdTypes.ASC,status,res,tid,self.latestChangeId)
    
    
    def ExecAbc(self,args,tid):
        status = ResTypes.OKY
        #res = False  
        history = self.GetHistory(tid)
        callId = self.qryEvaluator.Eval(args,history)
        res = self.callManager.AbortCall(callId, tid)
        self.AddToHistory(res,tid)
        return Res(CmdTypes.CAL,status,res,tid,self.latestChangeId)

    def ExecAdc(self,args,tid):
        status = ResTypes.OKY
        transaction = self.GetTransaction(tid)
        history = transaction.history
        res = self.Adc(*args)
        self.Val(res,tid) # direct evaluation after creation of constraint
        self.AddToHistory(res, tid)
        cid = self.AddLocalChange(tid,ChgTypes.ADC,args[1],args[2],args[4],None)
        res = [res.constraintType, res.target, res.feature, res.law, res.name, res.annotation]
        return Res(CmdTypes.ADC, status, res, tid, self.latestChangeId)

    def ExecRmc(self,args,tid):
        status = ResTypes.OKY
        transaction = self.GetTransaction(tid)
        history = transaction.history
        if args is None:
            self.logger.Info("validation: removing all constraints")
            constraint_target, constraint_feature, constraint_name = self.validationManager.RemoveAllConstraints()
            res = [constraint_target, constraint_feature, constraint_name]
        else:
            res = self.Rmc(args,tid)
            print("res: ",res)
            constraint_target = []
            constraint_feature = []
            constraint_name = []
            if not(isinstance(res,list)):
                res = [res]
            elif isinstance(res[0],list):
                res = res[0]
            for constraint in res:
                constraint_target.append(constraint.target)
                constraint_feature.append(constraint.feature)
                constraint_name.append(constraint.name)
        cid = self.AddLocalChange(tid,ChgTypes.RMC,constraint_target,constraint_feature,constraint_name,None)
        result = []
        if not(isinstance(res, list)):
            res = [res]
        for constraint in res:
            try:
                result.append([constraint.constraintType, constraint.target, constraint.feature, constraint.law, constraint.name, constraint.annotation])
            except:
                result.append(constraint)
        return Res(CmdTypes.RMC, status, result, tid, self.latestChangeId)

    def ExecVal(self,args,tid):
        obj1_new = []
        status = ResTypes.OKY
        transaction = self.GetTransaction(tid)
        history = transaction.history
        if args is None:
            res, obj1, obj2 = self.validationManager.ValidateAll(tid)
            for obj in obj1:
                obj1_new.append(obj.v)
        else:
            res, obj1, obj2 = self.Val(args, tid)
        self.AddToHistory(res, tid)
        return Res(CmdTypes.VAL, status, res, tid, self.latestChangeId)

    def ExecGac(self,args,tid):
        status = ResTypes.OKY
        transaction = self.GetTransaction(tid)
        history = transaction.history
        res = self.Gac()
        self.AddToHistory(res, tid)
        result = []
        for constraint in res:
            try:
                result.append([constraint.constraintType, constraint.target, constraint.feature, constraint.law, constraint.name, constraint.annotation])
            except:
                result.append(constraint)
        return Res(CmdTypes.GAC, status, result, tid, self.latestChangeId)
    
    #@Override
    def IsEventDesired(self,evt,callback,eventTypes,context,sessionId):
        #extend decision based on the session
        if(sessionId):
            if(sessionId in self.sessions):
                eventType = evt.evt
                typeKeys = self.sessions[sessionId].events[eventType]
                eventKey = evt.k
                isEventOk = (eventKey in typeKeys) or ('*' in typeKeys)
                return isEventOk and (evt.evt in eventTypes)
        else:
            return super().IsEventDesired(evt,callback,eventTypes,context,sessionId)
        
    
    '''
    MAINTENANCE METHODS
    '''     
   
    def OnCallManagerEvent(self,evts,source):
        self.NotifyObservers(evts,self) #forward all events of the call manager to the the own listeners
        
    def OnCallMdbAccessorEvent(self,evts,source):
        self.NotifyObservers(evts,self) #forward all events of the call manager to the the own listeners
        
    '''
    PRIVATE METHODS
    '''
    
    def StartTransaction(self,sessionId=None):
        self.mdbAccessor.Lock() #lock the db for this transaction
        self.latestTransactionId+=1
        tid = self.latestTransactionId
        self.mdbAccessor.latestTransactionId = self.latestTransactionId
        self.logger.Log('transaction','Transaction %d started.'%(tid))
        transaction = Transaction(tid)
        self.transactions[tid] = transaction
        #init session if required
        if(sessionId):
            self.InitSession(transaction, sessionId)
        return tid
    
    def EndTransaction(self,res,tid):
        #obtain the changes done during the transaction
        #transactionChanges = self.tempChangesForTransaction.pop(tid)
        
        transaction = self.transactions.pop(tid)
        transactionChanges = transaction.changes
        
        if(res.s == ResTypes.ERR):
            reason = res.v
            self.logger.Warn('Transaction %d failed: %s'%(tid,reason))
            nChanges = len(transactionChanges)
            if(0<nChanges):
                self.logger.Warn('Transaction %d: Rolling back %d changes.'%(tid,nChanges))
                self.RollbackChanges(transactionChanges)
        else:
            if(0<len(transactionChanges)):
                self.PersistChanges(transactionChanges)
        
        #delete the history
        self.mdbAccessor.Release()
        #self.tempHistoryForTransaction.pop(tid)
        self.logger.Log('transaction','Transaction %d ended.'%(tid))
        
    def GetTransaction(self,tid):
        transaction = None
        try:
            transaction = self.transactions[tid]
        except KeyError:
            raise EoqError(0,"Invalid transaction id %d"%tid)
        return transaction
    
    def InitSession(self,transaction,sessionId):
        session = None
        try:
            session = self.sessions[sessionId]
        except KeyError:
            #if not existing, create a new one
            session = self.CreateNewSession(sessionId=sessionId)
        #set the current transaction to the session
        transaction.session = session
        return session
    
    def CreateNewSession(self,sessionId=None,inheritSession=None,isTemp=False):
        if(None==sessionId):
            sessionId = str(uuid4())
            while(sessionId in self.sessions):
                sessionId = str(uuid4())
        elif(sessionId in self.sessions):
            raise EoqError('Session with id %s does already exist.'%(sessionId))
        self.sessionCount += 1
        sessionNumber = self.sessionCount
        session = Session(sessionId,sessionNumber)
        session.isTemp = isTemp
        if(inheritSession):
            session.user = inheritSession.user #inherit the user and access rights
        self.sessions[sessionId] = session
        return session
    
    def CloseTempSession(self,sessionId):
        if(sessionId in self.sessions):
            session = self.sessions[sessionId]
            if(session.isTemp):
                self.sessions.pop(sessionId) #delete the session
            else:
                raise EoqError(0,'Session %s is not a temp session.'%(sessionId))
        else:
            raise EoqError(0,'Unknown session %s'%(sessionId))
        return
        
    def GetHistory(self,tid):
        return self.GetTransaction(tid).history
        
    def AddToHistory(self,value,tid):
        self.GetTransaction(tid).history.append(value)
        
    def AddLocalChange(self,tid,ctype,target,feature,newVal,oldVal,oldOwner=None,oldFeature=None,oldIndex=None):
        transaction = self.GetTransaction(tid)
        cid = self.latestChangeId + len(transaction.changes)#+1
        user = ''
        sessionNumber = 0
        if transaction.session:
            user = transaction.session.user
            sessionNumber = transaction.session.sessionNumber
        chg = ChgEvt(cid,ctype,target,feature,newVal,oldVal,oldOwner,oldFeature,oldIndex,tid,user,sessionNumber) #change id is defined later if the command is done
        self.GetTransaction(tid).changes.append(chg)
        return cid
        
    def PersistChanges(self,changes):
        #notify observers on changes
        for chg in changes:
            #if the maximum of saved changes is exceeded begin deleting entries
            if(self.latestChangeId - self.earliestChangeId > self.maxChanges):
                self.changes.pop(self.earliestChangeId)
                self.earliestChangeId +=1
            #add each change of the current transaction to the change list
            self.changes[self.latestChangeId] = chg
            if(self.latestChangeId!=chg.a[0]):
                raise EoqError(0,'Inconsistency in change detected list.')
            self.latestChangeId+=1
            #log the changes if logging is enabled
            self.logger.Log('change',"Change (%d): %s t: %s, f: %s, n: %s, [was: v:%s, o: %s, f:%s, i:%s] (tid:%d, u:%s, sn:%d)"%(chg.a[0],chg.a[1],chg.a[2],chg.a[3],chg.a[4],chg.a[5],chg.a[6],chg.a[7],chg.a[8],chg.a[9],chg.a[10],chg.a[11]))
        self.NotifyObservers(changes,self)
        
    
        
    def RollbackChanges(self,changes):
        #revert local changes for multicommand transactions
        try:
            for chg in reversed(changes):
                self.logger.Warn("ROLLBACK: %s t: %s, f: %s, n: %s, o:%s"%(chg.a[1],chg.a[2],chg.a[3],chg.a[4],chg.a[5]))
                ctype = chg.a[1]
                if(ctype == ChgTypes.SET):
                    self.RollbackSet(chg)
                elif(ctype == ChgTypes.ADD):
                    self.RollbackAdd(chg)
                elif(ctype == ChgTypes.REM):
                    self.RollbackRem(chg)
                elif(ctype == ChgTypes.MOV):
                    self.RollbackMov(chg)
        except Exception as e:
            self.logger.Error("Error during rollback: %s"%(str(e)))  
            traceback.print_exc()
            
    def RollbackSet(self,chg):
        target = chg.a[2]
        feature = chg.a[3]
        newValue = chg.a[4]
        oldValue = chg.a[5]
        oldOwner = chg.a[6]
        oldFeature = chg.a[7]
        oldIndex = chg.a[8]
        #reset the old value
        self.mdbAccessor.Set(target,feature,oldValue)
        #rebuild the old values containment if it was existing before
        if(oldOwner):
            if(None==oldIndex):
                self.mdbAccessor.Set(oldOwner,oldFeature,newValue)
            else:
                self.mdbAccessor.Add(oldOwner,oldFeature,newValue)
                self.mdbAccessor.Move(newValue,oldIndex)
               
    def RollbackAdd(self,chg):
        target = chg.a[2]
        feature = chg.a[3]
        newValue = chg.a[4]
        #oldValue = chg.a[5]
        oldOwner = chg.a[6]
        oldFeature = chg.a[7]
        oldIndex = chg.a[8]
        #reset the old 
        #rebuild the old values containment if it was existing before
        if(oldOwner):
            if(None==oldIndex):
                self.mdbAccessor.Set(oldOwner,oldFeature,newValue)
            else:
                self.mdbAccessor.Add(oldOwner,oldFeature,newValue)
                self.mdbAccessor.Move(newValue,oldIndex)
        else: #was a new or free object
            self.mdbAccessor.Remove(target,feature,newValue)
            
    def RollbackRem(self,chg):
        target = chg.a[2]
        feature = chg.a[3]
        newValue = chg.a[4]
        #oldValue = chg.a[5]
        #oldOwner = chg.a[6]
        #oldFeature = chg.a[7]
        oldIndex = chg.a[8]
        #reset the old 
        #rebuild the old values containment if it was existing before
        self.mdbAccessor.Add(target,feature,newValue)
        self.mdbAccessor.Move(newValue,oldIndex)
            
    def RollbackMov(self,chg):
        target = chg.a[2]
        #feature = chg.a[3]
        #newValue = chg.a[4]
        #oldValue = chg.a[5]
        #oldOwner = chg.a[6]
        #oldFeature = chg.a[7]
        oldIndex = chg.a[8]
        #reset the old 
        #rebuild the old values containment if it was existing before
        self.mdbAccessor.Move(target,oldIndex)
        
    def Obs(self,sessionId,eventType,eventKey):
        session = self.sessions[sessionId]
        if((eventType not in ALL_EVENT_TYPES) and not(eventType == '*')):
            raise EoqError(0,'Unknown event type: %s'%(eventType))
        if('*' == eventType): #wildcard event
            for e in session.events:
                session.events[e][eventKey] = True
        else: #no wildcard event
            session.events[eventType][eventKey] = True
              
    def Ubs(self,sessionId,eventType,eventKey):
        session = self.sessions[sessionId]
        if((eventType not in ALL_EVENT_TYPES) and not(eventType == '*')):
            raise EoqError(0,'Unknown event type: %s'%(eventType))
        if('*' == eventType): #wildcard event
            if('*' == eventKey): #if key is a wildcard, all events are removed
                for e in session.events:
                    session.events[e].clear()
            else: #else remove the specific key
                for e in session.events:
                    session.events[e].pop(eventKey)
        else: #no wildcard event
            if('*' == eventKey):
                session.events[eventType].clear()
            else:
                session.events[eventType].pop(eventKey)

    def Adc(self,constraintType,target,feature,law,name,annotation):
        return self.validationManager.AddConstraint(constraintType,target,feature,law,name,annotation)

    def Rmc(self,constraint,tid):
        try:
            # TODO redo after constraints are properly integrated into model
            # hacky: convert history object to possible constraint
            history = self.GetHistory(tid)
            constraint2 = self.qryEvaluator.Eval(constraint,history)
        except:
            constraint2 = None
        if not(isinstance(constraint2,list)):
            constraint2 = [constraint2]
        elif isinstance(constraint2[0],list):
            constraint2 = constraint2[0]
        for con in constraint2:
            if self.validationManager.IsConstraint(constraint):
                # this only works if constraint is a native Constraint instance
                self.validationManager.RemoveConstraint(constraint)
            elif self.validationManager.IsConstraint(con):
                # therefore we try to convert a EOQ His object to its "value" above
                self.validationManager.RemoveConstraint(con)
        return constraint2

    def Val(self, obj, tid):
        """
            can evaluate either obj that are constraints (case 1 and 2) or obj that are targets (case 3)
        """
        result = []
        try:
            # TODO redo after constraints are properly integrated into model
            # hacky: convert history object to possible constraint
            #cmd = self.textSerializer.Des("GET "+self.textSerializer.Ser(obj))
            #obj2 = self.validationManager._RunCmd(cmd, tid).v
            history = self.GetHistory(tid)
            obj2 = self.qryEvaluator.Eval(obj,history)
        except:
            obj2 = None
        if not(isinstance(obj2,list)):
            obj2 = [obj2]
        elif isinstance(obj2[0],list):
            obj2 = obj2[0]
        for obj3 in obj2:
            if self.validationManager.IsConstraint(obj):
                # this only works if obj is a native Constraint instance
                res = self.validationManager.ValidateConstraint(obj, tid)
            elif self.validationManager.IsConstraint(obj3):
                # therefore we try to convert a EOQ His object to its "value" above
                res = self.validationManager.ValidateConstraint(obj3, tid)[0]
            else:
                # otherwise it is not a constraint but a valid target
                res = self.validationManager.ValidateTarget(obj, tid)
            result.append(res)
        return result, obj, obj2

    def Gac(self):
        return self.validationManager.GetConstraints()
        
    def ChangeConstrainttoList(self, res):
        for index in range(len(res)):
            if isinstance(res[index],list):
                self.ChangeConstrainttoList(res[index])
            elif self.validationManager.IsConstraint(res[index]):
                res[index] = [res[index].constraintType, res[index].target, res[index].feature, res[index].law, res[index].name, res[index].annotation]
            else:
                res[index] = res[index]
        return res




'''
Example script demonstrating the use of the Essential Object Queries to read and modify ECORE models. 

Bjoern Annighoefer 2018
'''
from ..model import DomainA
from ..domains import SimplePythonDomainWrapper

import fileinput

#ERROR HANLDERS
class EoqCsvError(Exception):
    pass

class EoqCsvReadError(EoqCsvError):
    def __init__(self, errortype, message, eoqstring):
        self.errortype = errortype
        self.message = message
        self.eoqstring = eoqstring
        
           
#MAIN CLASS
class CsvConnector: 
    def __init__(self,eoq : DomainA):
        self.elementIdLOT = {} #lookup table for elements ids
        self.elementLOT = {} #lookup table for elements
        self.idAttributeLOT = {} #lookup table id attributes
        self.dc = ';' #delimiter char
        self.qc = '"' #quote char
        self.sc = ',' #sub delimiter sign
        self.eoqOriginal = eoq
        self.eoq = SimplePythonDomainWrapper(eoq) #Hack to ensure backwards compatibility. TODO: Change to command interface later
     
    # splits a string by the given delimiter. but respects quote characters (qc) 
    # for encapsulating strings containing the delimiter characters (dc)
    def _csvSplit(self,txt):
        QUOTE_START = 1
        DELIM_START = 2
        state = DELIM_START
        b = 0 #begin of current substring
        e = 0 #end of current substring
        
        segments = []
        
        n = len(txt)
        for i in range(n):
            c = txt[i]
            if(DELIM_START == state):
                if(self.dc == c):
                    segment = txt[b:e]
                    segments.append(segment)
                    b = i+1
                elif(self.qc == c):
                    b = i+1
                    state = QUOTE_START
                else:
                    e = i+1
            elif(QUOTE_START == state):
                if(self.qc == c):
                    state = DELIM_START
                else:
                    e = i+1
        
        if(QUOTE_START == state):
            raise EoqCsvReadError('CVS split','Unbalanced quote starting at position %d.'%(b),txt)
        
        segment = txt[b:n]
        segments.append(segment)
    
        return segments;
    
    def _tanslateArray(self,strArray,translationDictionary):
        n = len(strArray)
        for i in range(n):
            if strArray[i] in translationDictionary:
                strArray[i] = translationDictionary[strArray[i]]
        return strArray
    
    def _writeLine(self,lineArray,dc,file):
        n = len(lineArray)
        if(n>0):
            print(self._csvOutputSegment(lineArray[0]),end="",file=self.file)
        for i in range(1,n):
            print("%c%s"%(dc,self._csvOutputSegment(lineArray[i])),end="",file=self.file)
        print("",file=self.file)
            
    
    #adds quote to a segment if it includes forbidden chars like a delimiter
    def _csvOutputSegment(self,segmentTxt):
        if(self.dc in segmentTxt or ' ' in segmentTxt):
            return '%c%s%c'%(self.qc,segmentTxt,self.qc)
        return segmentTxt #do nothing by default
    
    def getIdAttribute(self,element):
        idAttribute = None
        #clazz = element.eClass
        clazz = self.eoq.Retrieve(element,'(EObject:1)/eClass')
        if(clazz in self.idAttributeLOT):
            idAttribute = self.idAttributeLOT[clazz]
        else:
            idAttribute = self.eoq.Retrieve(clazz,'(EClass:1)/eAttributes{iD=true}')
            if(None == idAttribute):
                idAttribute = self.eoq.Retrieve(clazz,'(EClass:_)/eSuperTypes<-1>/eAttributes{iD=true}')
                if(len(idAttribute)>0):
                    idAttribute = idAttribute[0] #take only the first one. Multiple id attributes are not supported.
        self.idAttributeLOT[clazz] = idAttribute
        return idAttribute
          
    #return an element id
    def getElementId(self,element):
        if(element.v in self.elementIdLOT):
            return self.elementIdLOT[element.v]
        
        newElementId = None
        idAttribute = self.getIdAttribute(element);
        idAttributeName = self.eoq.Retrieve(idAttribute,'(EAttribute:1)/name')
        if(idAttribute):
            idValue = self.eoq.Retrieve(element,'(EObject:1)/%s'%(idAttributeName))
            if(idValue):
                if(idValue in self.elementLOT):
                    existingElement = self.elementLOT[idValue]
                    print('WARNING: %s seems to have the non unique id %s. %s has the same id. Using the path instead.'%(element,idValue,existingElement))
                else:
                    newElementId = str(idValue)
                    self.elementLOT[newElementId] = element.v
        if(None == newElementId):
            #no id available, so generate one by the path
            containerId = self.getContainerId(element)
            containedPath = self.getContainingPath(element,1)
            newElementId = "%s/%s"%(containerId,containedPath)
                
        self.elementIdLOT[element.v] = newElementId
        return newElementId
    
    def getContainerId(self,element):
        container = self.eoq.Retrieve(element,'(EObject:1)@CONTAINER')
        if(container):
            return self.getElementId(container)
        return ''
    
        '''
    retrieves the path of an object
    depth is the number of anchestors to start from (use -1 to obtain the absolute path)
    '''
    def getContainingPath(self,obj,depth):
        path = self.eoq.Retrieve(obj,'(EObject:1)@CONTAININGFEATURENAME')
        index = self.eoq.Retrieve(obj,'(EObject:1)@INDEX')
        if(index>0):
            path += '[%d]'%(index)
                    
        depth = depth-1 #reduce depth if one level was resolved
        
        if(depth!=0):
            container = self.eoq.Retrieve(obj,'(EObject:1)@CONTAINER')
            path = "%s/%s"%(self.getContainingPath(container,depth),path)
        
        return path

    def storeClass(self,package,clazz,gcmRoot,ommitContainer,ommitContainingPath,attributeNameDictionary,classSectionBegin,classSectionEnd,ommitClassInfo):
        clazzName = self.eoq.Retrieve(clazz,'(EClass:1)/name')
        allInstances = self.eoq.Retrieve(gcmRoot,"(*:*)$%s"%(clazzName))
        #print all attribute names
        if(0 < len(allInstances)):
            #allAttributes = clazz.eAllAttributes()
            ownAttributes = self.eoq.Retrieve(clazz,'(EClass:*)/eAttributes')
            superTypeAttributes = self.eoq.Retrieve(clazz,'(EClass:_)/eSuperTypes<-1>/eAttributes')
            allAttributes = ownAttributes+superTypeAttributes
            #allReferences = clazz.eAllReferences()
            ownNonContainingReferences = self.eoq.Retrieve(clazz,'(EClass:*)$EReference{containment=false}')
            superTypeNonContainingReferences = self.eoq.Retrieve(clazz,'(EClass:_)/eSuperTypes<-1>$EReference{containment=false}') 
            allNonContainingReferences = ownNonContainingReferences + superTypeNonContainingReferences

            packageNsUri = self.eoq.Retrieve(package,'(EPackage:1)/nsURI')
            className = self.eoq.Retrieve(clazz,'(EClass:1)/name')
            
            headerArray = []
            
            print(classSectionBegin,end="",file=self.file)
            if(False==ommitClassInfo):
                print("#%c%s%c%s"%(self.dc,self._csvOutputSegment(packageNsUri),self.dc,self._csvOutputSegment(className)),file=self.file)
            if(False==ommitContainer):
                headerArray.append("container")
                #print("container",end="",file=self.file)
            if(False==ommitContainingPath):
                headerArray.append("containedPath")
            #print("container%ccontainedPath"%(self.dc),end="",file=self.file)
            for attribute in allAttributes:
                attributeName = self.eoq.Retrieve(attribute,'(EAttribute:1)/name')
                headerArray.append(attributeName)
                #print("%c%s"%(self.dc,self._csvOutputSegment(attributeName)),end="",file=self.file)
            for reference in allNonContainingReferences:
                referenceName = self.eoq.Retrieve(reference,'(EReference:1)/name')
                headerArray.append(referenceName)
                #print("%c%s"%(self.dc,self._csvOutputSegment(referenceName)),end="",file=self.file)
            #replace elements with names from the dictionary
            if(len(attributeNameDictionary)>0):
                headerArray = self._tanslateArray(headerArray,attributeNameDictionary)
            self._writeLine(headerArray,self.dc,self.file)
            #print("",file=self.file)
            #print all values
            for instance in allInstances:
                lineArray = []
                if(False==ommitContainer):
                    lineArray.append(self.getContainerId(instance))
                    #print("container",end="",file=self.file)
                if(False==ommitContainingPath):
#                     featureName = self.eoq.Retrieve(instance,'(EObject:1)@CONTAININGFEATURENAME')
#                     lineArray.append(featureName)
                    lineArray.append(self.getContainingPath(instance,1))
                #print('%s%c%s'%(self._csvOutputSegment(self.getContainerId(eoq,instance)),self.dc,self._csvOutputSegment(self.eoq.EoqPath(instance,1))),end="",file=self.file)
                for attribute in allAttributes:
                    attributeName = self.eoq.Retrieve(attribute,'(EAttribute:1)/name')
                    attributeValue = self.eoq.Retrieve(instance,'(%s:1)/%s'%(className,attributeName))
                    attributeStr = ''
                    if(None!=attributeValue):
                        attributeStr = str(attributeValue)
                    lineArray.append(attributeStr)
                    #print('%c%s'%(self.dc,self._csvOutputSegment(attributeStr)),end="",file=self.file)
                for reference in allNonContainingReferences:
                    referenceString = ''
                    referenceName = self.eoq.Retrieve(reference,'(EReference:1)/name')
                    referedElements = self.eoq.Retrieve(instance,'(%s:*)/%s'%(className,referenceName))
                    if(not isinstance(referedElements,list)):
                        referedElements = [referedElements]
                    nReferedElements = len(referedElements)
                    if(nReferedElements>0 and referedElements[0]):
                        referenceString += self.getElementId(referedElements[0])
                    for i in range(1,nReferedElements):
                        referenceString += self.sc+self.getElementId(referedElements[i])
                    lineArray.append(referenceString)
                    #print('%c%s'%(self.dc,self._csvOutputSegment(referenceString)),end="",file=self.file)
                self._writeLine(lineArray,self.dc,self.file)
                #print('',file=self.file)
            print(classSectionEnd,end="",file=self.file)
        

    def storePackage(self,package,gcmRoot,ommitRootElement,ommitContainer,ommitContainingPath,attributeNameDictionary,classSectionBegin,classSectionEnd,ommitClassInfo):
        #allClazzes = package.eClassifiers
        allClazzes = self.eoq.Retrieve(package,'(EPackage:_)/eClassifiers$EClass{abstract=false}')
        i = 0
        for clazz in allClazzes:
            if(i==0 and ommitRootElement):
                i += 1
            else:
                self.storeClass(package,clazz,gcmRoot,ommitContainer,ommitContainingPath,attributeNameDictionary,classSectionBegin,classSectionEnd,ommitClassInfo)
                i +=1
            
        #look for subpackages
        allSubpackages = self.eoq.Retrieve(package,'(EPackage:*)/eSubpackages')
        #allSubpackages = package.eSubpackages
        for subpackage in allSubpackages:
            self.storePackage(subpackage, gcmRoot,False,ommitContainer,ommitContainingPath,attributeNameDictionary,classSectionBegin,classSectionEnd,ommitClassInfo)
            
    def SaveAsCsv(self,modelRoot,filename,header='',ommitRootElement=False,ommitContainer=False,ommitContainingPath=False,attributeNameDictionary={},classSectionBegin='',classSectionEnd='',ommitClassInfo=False):
       
        #reset members
        self.elementIdLOT = {}
        self.elementLOT = {}
        self.idAttributeLOT = {}
        
        mainclass = self.eoq.Retrieve(modelRoot,'(*:1)/eClass')
        mainpackage = self.eoq.Retrieve(mainclass,'(*:1)/ePackage')
        
        with open(filename, 'w') as self.file:
            print(header,end="",file=self.file)
            self.storePackage(mainpackage,modelRoot,ommitRootElement,ommitContainer,ommitContainingPath,attributeNameDictionary,classSectionBegin,classSectionEnd,ommitClassInfo)
            
    def LoadFromCsv(self,filename):
        
        #CONSTANTS
        CLASS_OR_ELEMENT_LINE = 1
        FEATURE_LINE = 2
        
        #reset members
        self.elementIdLOT = {}
        self.elementLOT = {}
        self.idAttributeLOT = {}
        
        #the return value
        gcmRoot = None
        
        with fileinput.input(files=(filename)) as f:
            packageNsUri = None
            classname = None
            allAttributeNames = []
            allAttributeDataTypes = []
            
            
            featureColumns = []
            attributeColumns = []
            attributeColumnIndicies = []
            attributeColumnsType = []
            attributeDataTypeTranslators = []
            nonContainingReferenceColumns = []
            nonContainingReferenceColumnIndicies = []
            nFeatureColumns = 0
            clazz = None
            idAttribute = None
            idAttributeName = None
            
            state = CLASS_OR_ELEMENT_LINE
            
            n = 1;
            
            tmpElementInfos = []
            
            for line in f:
                #remove line endings
                #line = lineEndRemoverRe.sub(line,'')
                line = line.splitlines()[0]
                
                #create all classes 
                if(CLASS_OR_ELEMENT_LINE==state):
                    if('#'==line[0]):
                        segments = self._csvSplit(line)
                        if(3!=len(segments)):
                            raise EoqCsvReadError('LoadFromCsv 1','Found invalid package or class identifier in line %d'%(n),line)
                        packageNsUri = segments[1]
                        classname = segments[2]
                        dummyElement = self.eoq.Create(packageNsUri,classname) #this element is only for test purposes and to obtain the class information
                        if(None==dummyElement):
                            raise EoqCsvReadError('LoadFromCsv 2','Could not create instance of class %s from package %s line %d'%(classname,packageNsUri,n),line)
                        clazz = self.eoq.Retrieve(dummyElement,'(EObject:1)/eClass')
                        ownAttributes = self.eoq.Retrieve(clazz,'(EClass:*)/eAttributes')
                        superTypeAttributes = self.eoq.Retrieve(clazz,'(EClass:_)/eSuperTypes<-1>/eAttributes')
                        
                        allAttributes = ownAttributes+superTypeAttributes
                        allAttributeNames = self.eoq.Retrieve(allAttributes,'(*:*)/name')
                        allAttributeDataTypes = self.eoq.Retrieve(allAttributes,'(*:*)/eType/name')
                        ownNonContainingReferenceNames = self.eoq.Retrieve(clazz,'(EClass:*)$EReference{containment=false}/name')
                        superTypeNonContainingReferenceNames = self.eoq.Retrieve(clazz,'(EClass:*)/eSuperTypes<-1>$EReference{containment=false}/name')
                        allNonContainingReferenceNames = ownNonContainingReferenceNames+superTypeNonContainingReferenceNames
                        
                        idAttributes = self.eoq.Retrieve(allAttributes,'(*:_)$EAttribute{iD=true}')
                        if(len(idAttributes)==1):
                            idAttribute = idAttributes[0]
                            idAttributeName = self.eoq.Retrieve(idAttribute,'(EAttribute:1)/name')
                        
                        state = FEATURE_LINE
                    else:
                        #valueStrs = line.split(';')
                        valueStrs = self._csvSplit(line)
                        nValues = len(valueStrs)
                        if(nValues!=nFeatureColumns):
                            raise EoqCsvReadError('LoadFromCsv 2','Found %d valueStrs but expected %d in line %d'%(nValues,nFeatureColumns,n),line)
                        containerId = valueStrs[0]
                        containingPath = valueStrs[1]
                        element = self.eoq.Create(packageNsUri,classname)
                        
                        #obtain the elements id
                        elementId = '%s/%s'%(containerId,containingPath)
                        
                        #write attributes
                        j = 0
                        for i in attributeColumnIndicies:
                            if('' != valueStrs[i]): #Only store something, if a value is given
                                value = attributeDataTypeTranslators[j](valueStrs[i])
                                self.eoq.Update(element,'/%s'%(featureColumns[i]),value)
                                if(idAttributeName and idAttributeName==featureColumns[i]):
                                    if(valueStrs[i] in self.elementLOT):
                                        print('WARNING: Element %s seems to have a non-unique id %s. Using the path instead'%(element,valueStrs[i]))
                                    else:
                                        elementId = valueStrs[i]
                            j = j+1
                            
                        self.elementLOT[elementId] = element       
                                
                        #extract reference values 
                        referenceValues = []
                        for i in nonContainingReferenceColumnIndicies:
                            referenceValues.append(valueStrs[i])
                                
                        #split up containing path
                        containingReferenceIsMulti = False
                        containingReferenceIndex = -1
                        containingPathSegments = containingPath.split('[')
                        containingReferenceName = containingPathSegments[0]
                        if(1<len(containingPathSegments)):
                            containingReferenceIsMulti = True
                            containingPathSegments = containingPathSegments[1].split(']')
                            containingReferenceIndex = int(containingPathSegments[0])
                                                 
                        #store the new element and some additional information for later 
                        tmpElementInformation = {'elementId':elementId, 'containerId':containerId, 'containingPath':containingPath, 'referenceValues':referenceValues, 'containingReferenceName':containingReferenceName, 'containingReferenceIsMulti':containingReferenceIsMulti, 'containingReferenceIndex':containingReferenceIndex,'nonContainingReferenceColumns':nonContainingReferenceColumns}
                        tmpElementInfos.append(tmpElementInformation)   
                        
                elif(FEATURE_LINE==state):
                    #featureColumns = line.split(';')
                    featureColumns = self._csvSplit(line)
                    nFeatureColumns = len(featureColumns)
                    if(2>nFeatureColumns):
                        raise EoqCsvReadError('LoadFromCsv 3','Line %d has %columns, but the minimum is 2'%(n,nFeatureColumns),line)
                    attributeColumns = []
                    attributeColumnIndicies = []
                    attributeColumnsType = []
                    attributeDataTypeTranslators = []
                    nonContainingReferenceColumns = []
                    nonContainingReferenceColumnIndicies = []
                    for i in range(nFeatureColumns):
                        if(featureColumns[i] in allAttributeNames):
                            j = allAttributeNames.index(featureColumns[i])
                            attributeColumns.append(f)
                            attributeColumnIndicies.append(i)
                            attributeColumnsType.append(allAttributeDataTypes[j])
                            attributeDataType = allAttributeDataTypes[j]
                            #create converter functions
                            if('EBoolean'==attributeDataType):
                                dataTypeConverter = lambda s: True
                            elif('EString'==attributeDataType):
                                dataTypeConverter = lambda s: str(s)
                            elif('EDouble'==attributeDataType):
                                dataTypeConverter = lambda s: float(s)
                            elif('EFloat'==attributeDataType):
                                dataTypeConverter = lambda s: float(s)
                            elif('EInt'==attributeDataType):
                                dataTypeConverter = lambda s: int(s)
                            elif('EDate'==attributeDataType):
                                dataTypeConverter = lambda s: str(s)
                            else:
                                dataTypeConverter = lambda s: str(s)
                            #else:
                            #    raise EoqCsvReadError('LoadFromCsv 5','Attribute %s of class %s has an unsupported data type:%s'%(featureColumns[i],classname,attributeDataType),line)
                            attributeDataTypeTranslators.append(dataTypeConverter)
                        elif(featureColumns[i] in allNonContainingReferenceNames):
                            nonContainingReferenceColumns.append(featureColumns[i])
                            nonContainingReferenceColumnIndicies.append(i)
                            
                    state = CLASS_OR_ELEMENT_LINE
                n = n+1;    
                
            #elements must be sorted by their containment
            elementInformationSortKey = lambda x: x['containerId']+'/'+x['containingReferenceName']+'%10.d'%(x['containingReferenceIndex'])
            tmpElementInfos.sort(key=elementInformationSortKey)
                
            #resolve all references
            for elementInfo in tmpElementInfos:
                elementId = elementInfo['elementId']
                containerId = elementInfo['containerId']
                #containingPath = elementInfo['containingPath']
                containingReferenceName = elementInfo['containingReferenceName']
                containingReferenceIsMulti = elementInfo['containingReferenceIsMulti']
                nonContainingReferenceColumns = elementInfo['nonContainingReferenceColumns']
                referenceValues = elementInfo['referenceValues']
                #get the element
                element = self.elementLOT[elementId]
                
                #establish the containment
                if('' == containerId and ''== containingReferenceName):
                    gcmRoot = element
                else:
                    father = None
                    if('' == containerId):
                        father = gcmRoot
                    else:
                        father = self.elementLOT[containerId]
                    if(None==father):
                        raise EoqCsvReadError('LoadFromCsv 4','Element %s has a not resolvable container id %s'%(elementId,containerId),containerId)
                    if(containingReferenceIsMulti):
                        self.eoq.Update(father,'/%s[+]'%(containingReferenceName),element) #can use plus, because elements are sorted in the right order    
                    else:
                        self.eoq.Update(father,'/%s'%(containingReferenceName),element)             
                
                #resolve all other references
                nNonContainingReferences=len(nonContainingReferenceColumns)
                for i in range(nNonContainingReferences):
                    nonContainingReferenceName = nonContainingReferenceColumns[i]
                    if(''!=referenceValues[i]):
                        referredElementIds = referenceValues[i].split(',')
                        if(1<len(referredElementIds)):
                            for referredElementId in referredElementIds:
                                referredElement = self.elementLOT[referredElementId]
                                self.eoq.Update(element,'/%s[+]'%(nonContainingReferenceName),referredElement)
                        else:
                            referredElement = self.elementLOT[referredElementIds[0]]
                            self.eoq.Update(element,'/%s'%(nonContainingReferenceName),referredElement)
                            
        return gcmRoot
            
                
            
            
        
        





'''
 Bjoern Annighoefer 2019
 '''

from .frame import FrameTypes,Frame
from .framehandler import FrameHandler
from ..util import EoqError
from ..util.logger import NoLogging,LogLevels
from ..event import ALL_EVENT_TYPES

import traceback

class DomainFrameHandler(FrameHandler):
    def __init__(self,domain,logger=NoLogging()):
        super().__init__()
        self.domain = domain
        self.logger = logger
        #self.domain.Observe(self.OnDomainEvent)
    
    def Handle(self,frames,sessionId=None):
        for i in range(len(frames)):
            try:   
                if(FrameTypes.CMD == frames[i].eoq):
                    cmd = frames[i].dat
                    res = self.domain.RawDo(cmd,sessionId)
                    frames[i].dat = res
                    frames[i].eoq = FrameTypes.RES
                else: 
                    raise EoqError(0,"Can not handle frame type: %s"%(frames[i].eoq))
            except Exception as e:
                errorMsg = "Got invalid frame: %s"%(str(e))
                self.logger.Error(errorMsg)
                traceback.print_exc();
                frames[i].eoq = FrameTypes.ERR
                frames[i].dat = errorMsg
        return frames
    
    #@Override 
    def Gby(self,sessionId):
        self.domain.Gby(sessionId)
    
    #@Override 
    def Observe(self,callback,eventTypes=ALL_EVENT_TYPES,context=None,sessionId=None): #by default register for all events
        self.domain.Observe(callback,eventTypes,context,sessionId)
    
    #@Override    
    def Unobserve(self,callback,context=None):
        self.domain.Unobserve(callback,context)


'''
 2019 Bjoern Annighoefer
'''

#check for all actions in folder


from ..action import Action,ActionArg,ActionArgOption
from ..call import CallStatus,CallOptions
from ..callhandler import CallHandler
from ...serialization import JsonSerializer
from ...domain.multiprocessing import MultiprocessingQueueDomainClient,MultiprocessingQueueDomainHost
from ...command.command import Abc
from ...util import EoqError,NoLogging,DATE_TIME_STR_FORMAT

import sys
import os
import imp
import time
import traceback
from datetime import datetime

from multiprocessing import Process
import threading

from timeit import default_timer as timer 


class SyncWriteRedirector:
    def __init__(self,stream,callId,channelName,callManager,silent=False,eventsonly=False,bufferTime=0.0):
        self.stream = stream
        self.callId = callId
        self.callManager = callManager
        self.channelName = channelName
        self.silent = silent
        self.eventsonly = eventsonly
        self.bufferTime = bufferTime
        
        self.lastWriteTime = time.perf_counter()
        self.buffer = ''
        
    def Redirect(self):
        self.legacyWriteFcn = self.stream.write
        self.stream.write = self.write

    def Restore(self):
        self.flush()
        self.stream.write = self.legacyWriteFcn
     
    # with handlers   
    def __enter__(self):
        self.Redirect()
        
    def __exit__(self,exc_type,exc_val,exc_tb):
        self.Restore()
         
    # overwrite stream functions        
    def write(self,data,*args):
            currentWriteTime = time.perf_counter()
            try:
                if(self.legacyWriteFcn and not self.silent and not self.eventsonly):
                    self.legacyWriteFcn(data)
                if(isinstance(data, str) and 0<len(data) and not self.silent):
                    #data = data.encode('utf-8')
                    self.buffer+=data
                    if(currentWriteTime-self.lastWriteTime>self.bufferTime):
                        self.callManager.AddCallOutputRaw(self.callId,self.channelName,self.buffer)
                        self.buffer = ''
                        self.lastWriteTime = currentWriteTime
            except Exception as e:
                sys.stderr.write(str(e))
    def flush(self):
            try:
                if(0<len(self.buffer) and not self.silent):
                    self.callManager.AddCallOutputRaw(self.callId,self.channelName,self.buffer)
                    self.buffer = ''
            except:
                pass
        
# 
# class WriteRedirector:
#     def __init__(self,callId,channelName,callManager,legacyWriteFcn=None,bufferTime=0.0):
#         self.callId = callId
#         self.callManager = callManager
#         self.channelName = channelName
#         self.legacyWriteFcn = legacyWriteFcn
#         self.bufferTime = bufferTime
#         
#         self.lastWriteTime = time.perf_counter()
#         self.buffer = ''
# 
#     def write(self,data, *args):
#             currentWriteTime = time.perf_counter()
#             try:
#                 if(self.legacyWriteFcn):
#                     self.legacyWriteFcn(data)
#                 if(isinstance(data, str) and 0<len(data)):
#                     #data = data.encode('utf-8')
#                     self.buffer+=data
#                     if(currentWriteTime-self.lastWriteTime>self.bufferTime):
#                         self.callManager.AddCallOutput(self.callId,self.channelName,self.buffer)
#                         self.buffer = ''
#                         self.lastWriteTime = currentWriteTime
#             except Exception as e:
#                 print(str(e))
#     def flush(self):
#             try:
#                 if(0<len(self.buffer)):
#                     self.callManager.AddCallOutput(self.callId,self.channelName,self.buffer)
#                     self.buffer = ''
#             except:
#                 pass
#         
#     def getStdStreamCompatibleWriteRedirector(self):
#         return (lambda data, *args: self.write(data, *args))

def AsyncCallProcessMain(actiondir,actionname,callId,argStr,optsStr,cmdQueue,resQueue,evtQueue):
    serializer = JsonSerializer()
    #restore args
    args = serializer.Des(argStr)
    #restore opts
    opts = CallOptions()
    opts.FromArray(serializer.Des(optsStr))
    #connect to domain
    domain = MultiprocessingQueueDomainClient(cmdQueue,resQueue,evtQueue,callId)
    stdoutRedirector = SyncWriteRedirector(sys.stdout,callId,'STDOUT',domain.callManager,silent=opts.silent,eventsonly=opts.eventsonly)
    stdoutRedirector.Redirect()
    relpathname = actionname+".py"
    pythonfile = os.path.join(actiondir,relpathname)
    pathname, filename = os.path.split(pythonfile)
    actionfunctionname = filename[:-3]
    modulename = actionname
    file, modulepath, description = imp.find_module(actionfunctionname, [pathname])
    sys.path.append(pathname)
    actionmodule = imp.load_module(modulename, file, modulepath, description)
    actionFunction = getattr(actionmodule,actionfunctionname)
    #convert input values
    if(actionFunction):
        stime = timer()
        callStatusInfo = "Python action %s started %s "%(relpathname,datetime.now().strftime(DATE_TIME_STR_FORMAT)) #is extended later
        domain.callManager.ChangeCallStatus(callId,CallStatus.RUN,callStatusInfo)
        finalCallStatus = CallStatus.RUN #is changed later
        try:
            value = actionFunction(domain,*args)
            domain.callManager.SetCallValue(callId, value)
            etime = timer()
            finalCallStatus = CallStatus.FIN
            callStatusInfo += " and ended %s (elapsed time: %f s) "%(datetime.now().strftime(DATE_TIME_STR_FORMAT),etime-stime)
        except Exception as e:
            traceback.print_exc()#
            etime = timer()
            finalCallStatus = CallStatus.ERR
            callStatusInfo += " and failed %s (elapsed time: %f s): %s"%(datetime.now().strftime(DATE_TIME_STR_FORMAT),etime-stime,str(e))
        #make sure all output was printed to the domain
        stdoutRedirector.flush()
        stdoutRedirector.Restore()
        #finally change the call status to FINISHED
        domain.callManager.ChangeCallStatus(callId,finalCallStatus,callStatusInfo)
    domain.Stop() #stop the event listener thread of the domain
    
            
        
class RunningAction:
    #def __init__(self,actionCall,process,domainWrapper):
    def __init__(self,callId):
        self.started = False
        self.aborted = False
        self.callId = callId
        self.process = None
        self.domainWrapper = None
        
    def SetStarted(self,process,domainWrapper):
        self.started = True
        self.process = process
        self.domainWrapper = domainWrapper
        
    def HasStarted(self):
        return self.started;
    
    def IsAborted(self):
        return self.aborted;
    
    def Abort(self):
        self.aborted = True;
        
'''
    The external action handler
'''
class ExternalPyScriptHandler(CallHandler):
    
    def __init__(self,callManager,basedir='actions',logger=NoLogging()):
        #constants
        self.ANNOTATION_ESCAPE_SYMBOLS = [r'\:',r'\[',r'\]',r'\{',r'\}',r"\'",r'\"',r'\\']
        #parameters
        self.callManager = callManager
        self._basedir = basedir
        self.logger = logger
        
        self._actions = {} #name -> (action,handlerfunction)
        self._runningActions = {}
        #register default action for reloading action files        
        self.__ReloadActions()
        return
    
    def __ScanForActions(self):
    
        #find all python files in the action directory
        pythonfiles = []
        for root, dirs, files in os.walk(self._basedir, topdown=True):
            for file in files:
                if(file.endswith('.%s'%('py'))):
                    #relativeRoot = os.path.relpath(root, self._basedir)
                    #pythonfiles += [os.path.join(relativeRoot,file)] #2: omits the ./ at the beginning
                    pythonfiles += [os.path.join(root,file)] #2: omits the ./ at the beginning
        
        
        for pythonfile in pythonfiles:
            pathname, filename = os.path.split(pythonfile)
            if(os.path.isfile(os.path.join(pathname,'__init__.py'))):
                continue #skip any submodules from beeing loaded as an action
            actionfunctionname = filename[:-3] #without .py
            relpathname = os.path.relpath(pathname, self._basedir)
            modulename = None
            if(relpathname=="."): #it is the root folder
                modulename = actionfunctionname
            else: #any subfolder
                prefixfreepath = relpathname.replace('./','')
                packagename = prefixfreepath.replace('/','.')
                modulename = packagename+'.'+actionfunctionname
            actionname = modulename.replace('.','/')
            try: #prevent server crash by broken action scripts
                file, modulepath, description = imp.find_module(actionfunctionname, [pathname])
                sys.path.append(pathname)
                actionmodule = imp.load_module(modulename, file, modulepath, description)
                #check for a unique name
                if(actionfunctionname in self._actions):
                    self.logger.Warn("Found external action %s in file %s, but an action with the same name is already registered from file %s. Action names must be unique!"%(actionname,pythonfile,self._actions[actionname].filepath))
                    continue
                #check if the module contains a function with the same name
                if(actionfunctionname not in dir(actionmodule)):
                    self.logger.Warn("Skipped no-action %s in %s."%(actionname,pythonfile))
                    continue
                self.logger.Info("Found external action %s in file %s"%(actionname,pythonfile))
                #register the new action 
                
                actionFunction = getattr(actionmodule,actionfunctionname)
                actionArguments = self.__ParseActionArguments(actionname,actionFunction)
                actionResults = self.__ParseActionResults(actionname,actionFunction)
                actionDescription = actionmodule.__doc__
                actionTags = getattr(actionmodule, '__tags__', [])
                if actionTags:
                    self.logger.Info("Action has tags: %s"%(actionTags))
                actionCategory = ""
                action = Action(actionname,actionArguments,actionResults,actionDescription,actionCategory,actionTags)

                self._actions[actionname] = (action,actionFunction)
                #Action(actionname,pythonfile,actionFunction,actionArguments,actionResults,actionDescription)
            except Exception as e:
                self.logger.Error("Error loading external action %s from %s: %s"%(actionname,pythonfile,str(e)))
                traceback.print_exc(file=sys.stdout)
            
    def __ParseActionArguments(self,actionName,actionFunction):
        args = []
        nArguments = actionFunction.__code__.co_argcount
        functionVariables = actionFunction.__code__.co_varnames
        actionArguments = functionVariables[0:nArguments]
        actionArgumentTypeInfos = actionFunction.__annotations__
        
        if(nArguments==0):
            self.logger.Warn("Action %s has no argument. Actions must have at least one argument of type Domain."%(actionName))
        else:
            #look for the first argument. This must always be a domain
            argumentName = actionArguments[0]
            argument = None
            if(argumentName not in actionArgumentTypeInfos):
                self.logger.Warn("Argument 1 of action %s is not annotated assuming Domain as type."%(actionName))
            else:
                argumentAnnotation = actionArgumentTypeInfos[argumentName]
                if("Domain"!=argumentAnnotation):
                    self.logger.Warn("Argument 1 of action %s has type %s but expected is %s. This will probably not work."%(actionName,argumentAnnotation,"Domain"))
            #add all args after the first 
            for i in range(1,nArguments):
                argumentName = actionArguments[i]
                argument = None
                if(argumentName not in actionArgumentTypeInfos):
                    self.logger.Warn("Argument %s of action %s has no annotation. Assuming * as type"%(argumentName,actionName))
                    argument = ActionArg(argumentName,'*',1,1,'','',[])
                else:
                    argumentAnnotation = actionArgumentTypeInfos[argumentName]
                    [argumentType, argumentMin, argumentMax, description, default, choices] = self.__ParseArgumentAnnotation(argumentAnnotation)
                    argument = ActionArg(argumentName,argumentType, argumentMin, argumentMax, description, default, choices)
                args.append(argument)
        return args
        
    def __ParseActionResults(self,actionName,actionFunction):
        args = []
        nArguments = actionFunction.__code__.co_argcount
        functionVariables = actionFunction.__code__.co_varnames
        actionArguments = functionVariables[0:nArguments]
        actionArgumentTypeInfos = actionFunction.__annotations__
        
        argumentName = 'return' #'return' is the default key for return annotations in python
        argument = None
        if(argumentName not in actionArgumentTypeInfos):
            self.logger.Warn("Action %s has no return annotation. Assuming no return value"%(actionName))
        else:
            argumentAnnotation = actionArgumentTypeInfos[argumentName]
            if('' != argumentAnnotation):
                [argumentType, argumentMin, argumentMax, description, default, choices] = self.__ParseArgumentAnnotation(argumentAnnotation)
                args.append(ActionArg(argumentName,argumentType, argumentMin, argumentMax, description, default, choices))
        return args
    
    def __MaskEscapeArgumentAnnotation(self,annotation):
        maskedAnnotation = annotation
        for s in self.ANNOTATION_ESCAPE_SYMBOLS:
            maskedAnnotation = maskedAnnotation.replace(s,'x')
        return maskedAnnotation
        
    def __UnescapeArgumentAnnotation(self,annotation):
        
        maskedAnnotation = annotation
        for s in self.ANNOTATION_ESCAPE_SYMBOLS:
            maskedAnnotation = maskedAnnotation.replace(s,s[1])
        return maskedAnnotation
        
        
    def __ParseArgumentAnnotation(self,argumentAnnotation):
        #a annotation should look like this: <Type>[<multiplicity=0..1|*>]{<choice1>',<choice2>,...}=<default>:<Description>
        #all parameters except type are optional
        argumentType = ''
        argumentMin = 1
        argumentMax = 1
        description = ''
        default = ''
        choices = []
        
        #parse string
        nArgumentAnnotation = len(argumentAnnotation)+1
        maskedAnnotation = self.__MaskEscapeArgumentAnnotation(argumentAnnotation) #neglect all escaped symbols
        multiplicityStart = maskedAnnotation.find('[')
        choiceStart = maskedAnnotation.find('{')
        defaultStart = maskedAnnotation.find('=')
        descriptionStart = maskedAnnotation.find(':')
        # remove all escape sequences
        argumentAnnotation = self.__UnescapeArgumentAnnotation(argumentAnnotation)
        
        #multiplicity
        if(multiplicityStart > 0):
            multiplicityEnd = multiplicityStart+argumentAnnotation[multiplicityStart:].find(']')
            if(multiplicityEnd>0 and multiplicityEnd>multiplicityStart):
                multiplicityStr = argumentAnnotation[multiplicityStart+1:multiplicityEnd]
                if('..' in multiplicityStr):
                    multiplicity = multiplicityStr.split('..')
                    if(len(multiplicity)==2):
                        argumentMin = int(multiplicity[0])
                        if('*' == multiplicity[1]):
                            argumentMax = -1
                        else:
                            argumentMax = int(multiplicity[1])
                else:
                    if('*' == multiplicityStr):
                        argumentMin = 0
                        argumentMax = -1
                    else:
                        argumentMin = int(multiplicityStr)
                        argumentMax = int(multiplicityStr)
            else:
                self.logger.Warn("Malformed multiplicity in argument annotation %s"%(argumentAnnotation))

        #choice
        if(choiceStart > 0):
            choiceEnd = choiceStart+argumentAnnotation[choiceStart:].find('}')
            if(choiceEnd>0 and choiceEnd>choiceStart+1): #skip { and }
                choiceStr = argumentAnnotation[choiceStart+1:choiceEnd]
                choices = [ActionArgOption(v,'') for v in choiceStr.split(',')]
            else:
                self.logger.Warn("Malformed choice in argument annotation %s"%(argumentAnnotation))
        #default
        if(defaultStart > 0):
            defaultEnd = min([nArgumentAnnotation,
                              descriptionStart%nArgumentAnnotation])
            if(defaultEnd>0 and defaultEnd>defaultStart+1):
                default = argumentAnnotation[defaultStart+1:defaultEnd] #skip :' and '
            else:
                self.logger.Warn("Malformed default in argument annotation %s"%(argumentAnnotation))

        #description
        if(descriptionStart > 0):
            descriptionEnd = nArgumentAnnotation 
            if(descriptionEnd>0 and descriptionEnd>descriptionStart+1): #skip :
                description = argumentAnnotation[descriptionStart+1:descriptionEnd] #skip :
            else:
                self.logger.Warn("Malformed description in argument annotation %s"%(argumentAnnotation))


        #type
        typeStart = 0
        typeEnd = min([nArgumentAnnotation,
                       multiplicityStart%nArgumentAnnotation,
                       choiceStart%nArgumentAnnotation,
                       defaultStart%nArgumentAnnotation,
                       descriptionStart%nArgumentAnnotation])
        argumentType = argumentAnnotation[typeStart:typeEnd]                 
        return [argumentType, argumentMin, argumentMax, description, default, choices] 
        
    
    def __ReloadActions(self):
        #empty the list of actions
        for key in self._actions:
            self.callManager.UnregisterAction(key)
        self._actions.clear()
            
#         #register actions
#         actionname = 'reload-actions'
#         actionFunction = lambda d: self.__ReloadActions(d)
#         self._actions[actionname] = (Action(actionname,'',[],[],'Reloads the action files and their descriptions from the file system'),actionFunction)
#         #scan for file based actions
        self.__ScanForActions()
        #register all actions
        for item in self._actions.items():
            name = item[0]
            action = item[1][0] #action is the first element of the touble (action,handler)
            self.callManager.RegisterAction(name,action,self)
        return
        
    def HandleSync(self,name,callId,domain,args,opts,callManager):
        if(opts.timeout>0.0):
            self.logger.Warn('timeout=%f is not supported for CAL'%(opts.timeout))
        
        if(name not in self._actions):
            raise EoqError(0,'Action with name %s is unknown.'%(name))
        (action,actionFunction) = self._actions[name]
        
        callStatusInfo = "Python action %s started %s"%(name,datetime.now().strftime(DATE_TIME_STR_FORMAT))
        callManager.ChangeCallStatus(callId,CallStatus.RUN,callStatusInfo)
        finalCallStatus = CallStatus.RUN #is changed later
        value = None
        with SyncWriteRedirector(sys.stdout,callId,'STDOUT',callManager,silent=opts.silent,eventsonly=opts.eventsonly,bufferTime=0.0):
            stime = timer()
            try:
                value = actionFunction(domain,*args)
                finalCallStatus = CallStatus.FIN
                etime = timer()
                callStatusInfo += " and ended %s (elapsed time: %f s)"%(datetime.now().strftime(DATE_TIME_STR_FORMAT),etime-stime)
            except Exception as e: 
                traceback.print_exc()
                finalCallStatus = CallStatus.ERR
                etime = timer()
                callStatusInfo += " and failed %s (elapsed time: %f s): %s"%(datetime.now().strftime(DATE_TIME_STR_FORMAT),etime-stime,str(e))
                raise e #make sure the exception goes to the call manager
        callManager.ChangeCallStatus(callId,finalCallStatus,callStatusInfo)
        return value
    
    def HandleAsync(self,name,callId,domain,args,opts,callManager,sessionId=None):
        # AsyncCall(self, actionCall : ActionCall, transaction : Transaction):
        if(name not in self._actions):
            raise EoqError(0,'Action with name %s is unknown.'%(name))
        
        runningAction = RunningAction(callId)
        self._runningActions[callId] = runningAction
        
        observerThread = threading.Thread(target=self.__AsyncCallObserverThread, args=(name,runningAction,domain,args,opts,callManager,sessionId,callId,))
        observerThread.start()
        return

    def AbortCall(self, callId):
        #(self, actionCall : ActionCall, transaction : Transaction):
        if(callId not in self._runningActions):
            raise EoqError(0,'ExternalActionHandler: CallId %d is unknown'%(callId))
        #remove from the list of running processes
        runningAction = self._runningActions.pop(callId)
        #set an aborted status
        runningAction.Abort()
        
        if(runningAction.HasStarted()):
            #kill the process
            runningAction.process.terminate() 
        return True
    
    def __AsyncCallObserverThread(self,name,runningAction,domain,args,opts,callManager,sessionId,callId):
        #runningAction = self._runningActions[callId]
        serializer = JsonSerializer()
        argsStr = serializer.Ser(args)
        optsStr = serializer.Ser(opts.ToArray())
        hasTimout = (opts.timeout > 0.0)
        wasAborted = False
        domainWrapper = MultiprocessingQueueDomainHost(domain,callManager,sessionId=sessionId,callId=callId)
        proc = Process(target=AsyncCallProcessMain, args=(self._basedir,name,runningAction.callId,argsStr,optsStr,domainWrapper.cmdQueue,domainWrapper.resQueue,domainWrapper.evtQueue,))
#         runningAction = RunningAction(callId,proc,domainWrapper)
#         self._runningActions[callId] = runningAction
        sTime = timer()
    
        if(not runningAction.IsAborted()): #check if the call was aborted before the process was created
            domainWrapper.Start()
            proc.start()
            runningAction.SetStarted(proc,domainWrapper)
          
            while(proc.is_alive()): #wait until the thread has finished
                time.sleep(0.1)
                if(hasTimout):
                    cTime = timer()
                    pTime = cTime-sTime
                    if(not wasAborted and pTime > opts.timeout): 
                        #abort the process if it took to long
                        self.logger.Warn('Timeout of %f s reached for call %d. Aborting.'%(opts.timeout,callId))
                        domain.Do(Abc(callId))
                        wasAborted = True
                
            domainWrapper.Join()


'''
 2019 Bjoern Annighoefer
'''
from ..action.callmanager import CallManager
from ..action.call import CallOptions

from .call import CallTypes,CallStatus,Call,CALL_STATUS_FINAL_TYPES
from ..event import CstEvt,CvaEvt,OupEvt,ALL_EVENT_TYPES
from ..util import EoqError,DATE_TIME_STR_FORMAT

from datetime import datetime


'''
  CmdRunnerBaseCallManager
'''

class CmdRunnerBasedCallManager(CallManager):
    def __init__(self,cmdRunner,callId=-1):
        super().__init__(callId)
        self.cmdRunner = cmdRunner
        self.actionInfos = {} #'action name' -> (action,handler)
        self.currentCallId = 0 #only for async calls
        self.runningCalls = {} #callid -> (callback handler, abort handler, call status,)
    
    def RegisterAction(self,name,action,handler):
        self.actionInfos[name] = (action,handler)
        
    def UnregisterAction(self,name):
        self.actionInfos.pop(name)
        
    def GetAllActions(self):
        return [value[0] for key, value in sorted(self.actionInfos.items())]
    
    def FinalizeAsyncCall(self,call):
        #check if the action is terminated in case of an async action
        if(CallTypes.ASY == call.ctype and call.status in CALL_STATUS_FINAL_TYPES and call.hasValue):
            if(call.opts.autoobserve and call.parentSessionId):
                self.cmdRunner.Ubs(call.parentSessionId,'*',str(call.callId))
            self.cmdRunner.CloseTempSession(call.sessionId) #free the actions session
       
    def ChangeCallStatus(self,callId,status,info=''):
        if(callId in self.runningCalls):
            call = self.runningCalls[callId]
            if(call.status != status): #make sure this is really an update
                call.status = status
                self.NotifyObservers([CstEvt(callId,status,info)])
                self.FinalizeAsyncCall(call)
            
    def SetCallValue(self,callId,value):
        if(callId in self.runningCalls):
            call = self.runningCalls[callId]
            call.value = value
            call.hasValue = True
            self.NotifyObservers([CvaEvt(callId,value)],self)
            self.FinalizeAsyncCall(call)
            
    def AddCallOutputRaw(self,callId,channelName,data):
        if(callId in self.runningCalls):
            callInfo = self.runningCalls[callId]
            if(channelName in callInfo.outputs):
                callInfo.outputs[channelName] += data
            else:
                callInfo.outputs[channelName] = data
            self.NotifyObservers([OupEvt(callId,channelName,data)],self)
            
    def AddCallOutput(self,channelName,data):
        self.AddCallOutputRaw(self.callId,channelName,data)
            
    def RunCallSync(self,name,args,optsArr,tid,sessionId=None):
        #name = args[0]
        #obtain the handler
        info = self.actionInfos[name]
        action = info[0]
        handler = info[1]
        #check if the arguments are appropriate
        nActionArgs = len(args)
        nExpectedArgs = len(action.args)
        if(nActionArgs!=nExpectedArgs):
            raise EoqError(0,"Action %s requires %s arguments, but got %d."%(name,nExpectedArgs,nActionArgs))
        #parse options
        opts = CallOptions()
        opts.FromArray(optsArr)
        #call the handler
        domain = CallDomain(self.cmdRunner,tid) #create a private domain to be used within the action
        callInfo = self.CreateNewCall(domain,name,args,opts,handler,CallTypes.SYN,sessionId)
        
        if(opts.autoobserve and sessionId):
            self.cmdRunner.Obs(sessionId,'*',str(callInfo.callId))
        value = handler.HandleSync(name,callInfo.callId,domain,args,opts,self)
        if(opts.autoobserve and sessionId):
            self.cmdRunner.Ubs(sessionId,'*',str(callInfo.callId))
        callOutputs = [[i[0],i[1]] for i in callInfo.outputs.items()] #dict needs to be removed
        result = [callInfo.callId,value,callInfo.status,callOutputs]
        return result
    
    def RunCallAsync(self,name,args,optsArr,tid,sessionId=None,parentSessionId=None):
        info = self.actionInfos[name]
        action = info[0]
        handler = info[1]
        #check if the arguments are appropriate
        nActionArgs = len(args)
        nExpectedArgs = len(action.args)
        if(nActionArgs!=nExpectedArgs):
            raise EoqError(0,"Action %s requires %s arguments, but got %d."%(name,nExpectedArgs,nActionArgs))
        #parse options
        opts = CallOptions()
        opts.FromArray(optsArr)
        #call the handler
        domain = AsyncCallDomain(self.cmdRunner) #create a private domain to be used within the action
        callInfo = self.CreateNewCall(domain,name,args,opts,handler,CallTypes.ASY,sessionId,parentSessionId)
        
        if(opts.autoobserve and parentSessionId):
            self.cmdRunner.Obs(parentSessionId,'*',str(callInfo.callId))
        handler.HandleAsync(name,callInfo.callId,domain,args,opts,self,sessionId=sessionId)
        result = callInfo.callId
        return result
    
    
    def AbortCall(self,callId,tid):
        try:
            callInfo = self.runningCalls[callId]
            if(callInfo.ctype == CallTypes.SYN):
                raise EoqError(0,"Call %d is synchronous. Only asynchronous calls can be aborted."%(callId))
            if(callInfo.status in CALL_STATUS_FINAL_TYPES):
                raise EoqError(0,"Call %d is not running any more."%(callId))
            result = callInfo.handler.AbortCall(callId)
            if(result):
                callInfo.endTime = datetime.now()
                callStatusInfo = "Action %s started %s was aborted at %s (elapsed time: %d s) "%(
                    callInfo.name,
                    callInfo.startTime.strftime(DATE_TIME_STR_FORMAT),
                    callInfo.endTime.strftime(DATE_TIME_STR_FORMAT),
                    (callInfo.endTime-callInfo.startTime).total_seconds()
                    )
                self.ChangeCallStatus(callId, CallStatus.ABO,callStatusInfo)
        except KeyError:
            raise EoqError(0,"Call id %d is unknown."%(callId))
        return result
    
    '''
     PRIVATE METHODS
    '''
   
    def CreateNewCall(self,domain,name,args,opts,handler,ctype,sessionId=None,parentSessionId=None):
        callId = self.currentCallId
        self.currentCallId+=1
        callInfo = Call(callId,domain,name,args,opts,handler,ctype,sessionId,parentSessionId=parentSessionId)
        callInfo.startTime = datetime.now()
        self.runningCalls[callId] = callInfo
        return callInfo
    
'''
    A domain that wraps a specific transaction id
    Is used internally in the command runner for executing actions
'''   
        
from ..domain.domain import Domain
        
class CallDomain(Domain):
    def __init__(self,cmdRunner,tid):
        self.cmdRunner = cmdRunner
        self.tid = tid
       
    #@override 
    def RawDo(self,cmd,sessionId=None):
        # TODO: fix sessionId to do something
        res = self.cmdRunner.ExecOnTransaction(cmd,self.tid)
        return res
    
class AsyncCallDomain(Domain):
    def __init__(self,cmdRunner):
        super().__init__()
        self.cmdRunner = cmdRunner
       
    #@override 
    def RawDo(self,cmd,sessionId=None):
        res = self.cmdRunner.Exec(cmd,sessionId=sessionId)
        return res
    
    #Override the event provid methods since the sole event provider shall be the cmd runner
    #@Override
    def Observe(self,callback,eventTypes=ALL_EVENT_TYPES,context=None,sessionId=None): #by default register for all events
        self.cmdRunner.Observe(callback,eventTypes,context,sessionId)
    
    #@Override    
    def Unobserve(self,callback,context=None):
        self.cmdRunner.Unobserve(callback,context)
        
    #@Override
    def NotifyObservers(self,evts,excludedCallback=None,excludedContext=None):
        self.cmdRunner.NotifyObservers(evts,excludedCallback,excludedContext)



'''
 An EOQ protocol communication frame
 Bjoern Annighoefer 2019
 '''

class FrameTypes:
        CMD = "CMD" #Command
        RES = "RES" #Result
        CHG = "CHG" #Change
        EVT = "EVT" #Event: Change, call output
        ERR = "ERR" #Error
        
class Frame:
    def __init__(self,ftype,uid,data,version=210):
        self.eoq = ftype
        self.ver = version
        self.uid = uid
        self.dat = data
        
        


from pyecore.resources import global_registry
from .workspacemdbmodel import getEClassifier, eClassifiers
from .workspacemdbmodel import name, nsURI, nsPrefix, eClass
from .workspacemdbmodel import PathElementA, Workspace, Directory, FileResourceA, FileResourceTypesE, ModelResource, TextResource, RawResource, XmlResource

from pyecore.ecore import EObject

from . import workspacemdbmodel

__all__ = ['PathElementA', 'Workspace', 'Directory', 'FileResourceA',
           'FileResourceTypesE', 'ModelResource', 'TextResource', 'RawResource', 'XmlResource']

eSubpackages = []
eSuperPackage = None
workspacemdbmodel.eSubpackages = eSubpackages
workspacemdbmodel.eSuperPackage = eSuperPackage

Directory.subdirectories.eType = Directory
Directory.resources.eType = FileResourceA
ModelResource.contents.eType = EObject
XmlResource.document.eType = EObject

otherClassifiers = [FileResourceTypesE]

for classif in otherClassifiers:
    eClassifiers[classif.name] = classif
    classif.ePackage = eClass

for classif in eClassifiers.values():
    eClass.eClassifiers.append(classif.eClass)

for subpack in eSubpackages:
    eClass.eSubpackages.append(subpack.eClass)

register_packages = [workspacemdbmodel] + eSubpackages
for pack in register_packages:
    global_registry[pack.nsURI] = pack



'''
 Bjoern Annighoefer 2019
'''


from ..mdb import Mdb
from ...util import NoLogging,EoqError

from pyecore.resources import ResourceSet, URI
from pyecore.ecore import EPackage

from timeit import default_timer
from threading import Timer

class PyEcoreSingleFileMdb(Mdb):
    def __init__(self,modelfile,metamodelfile=None,saveTimeout=1.0,logger=NoLogging(),autoload=True):
        self.modelfile = modelfile
        self.metamodelfile = metamodelfile
        self.isLoaded = False
        self.rset = None
        self.modelroot = None
        
        self.logger = logger
        
        self.metamodelresource = None
        self.modelresource = None
        self.isDirty = False #indicates if the file needs to be saved
        self.saveTimeout = saveTimeout #This time is waited for any model modifications and than autosafe is called 
        self.saveTimer = None
        
        if(autoload):
            self.Load()
            
    def __del__(self):
        #make sure any changes are saved
        #self.logger.Info("Closing single %s ..."%(self.modelfile)) #makes problems when main thread has ended already
        
        self.__StopDelayedSaving()
        self.__Save()
        
    def Root(self):
        return self.modelroot
    
    def Class(self,packageName,className):
        clazz = None
        try:
            pack = self.rset.metamodel_registry[packageName]
            clazz = pack.getEClassifier(className)
        except Exception as e:
            raise EoqError(0,"Could not create class %s:%s: %s"%(packageName,className,str(e)))
        return clazz
        
    def Load(self,):
        self.logger.Info("Opening mdb for %s... "%(self.modelfile))
        start = default_timer()
        if(not self.isLoaded):
            self.rset = ResourceSet()
            if(self.metamodelfile):
                self.metamodelresource = self.rset.get_resource(URI(self.metamodelfile)) 
                metamodelroot = self.metamodelresource.contents[0]
                self.rset.metamodel_registry[metamodelroot.nsURI] = metamodelroot
                for child in metamodelroot.eAllContents():
                        if(isinstance(child,EPackage)):
                            self.rset.metamodel_registry[child.nsURI] = child
            self.modelresource = self.rset.get_resource(URI(self.modelfile)) 
            self.modelroot = self.modelresource.contents[0]  # We get the root (an EPackage here)
        end = default_timer()
        self.logger.Info("ok (%f s)"%(end-start))
            
    def Save(self):
        #stop the current save action 
        self.__StopDelayedSaving()
        if(self.isDirty):
            if(self.saveTimeout > 0):
                self.__StartDelayedSaving()
            else:
                self.__Save() #skip threading for a 0 timeout
        
    def __StartDelayedSaving(self):
        self.saveTimer = Timer(self.saveTimeout,self.__DelayedSaving)
        self.saveTimer.start()
        
    def __StopDelayedSaving(self):
        if(self.saveTimer):
            self.saveTimer.cancel()
            self.saveTimer = None
            
    def __DelayedSaving(self):
        self.__Save()
        self.saveTimer = None
        
    def __Save(self):
        if(self.isDirty):
            self.__SaveResource()
            self.isDirty = False
        
    def __SaveResource(self):
        self.logger.Info("Saving %s ..."%(self.modelfile))
        start = default_timer()
        self.modelresource.save()
        self.isDirty = False
        end = default_timer()
        self.logger.Info("ok (%f s)"%(end-start))
    
    def SetObjectDirty(self,obj):
        #quit any previous saving timeout, this prevents doing the save during heavy DB interactions
        self.__StopDelayedSaving()
        #mark it dirty
        self.isDirty = True
    


"""Definition of meta model 'workspacemdbmodel'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *


name = 'workspacemdbmodel'
nsURI = 'http://www.eoq.de/workspacemdbmodel/v1.0'
nsPrefix = 'de.eoq'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)
FileResourceTypesE = EEnum('FileResourceTypesE', literals=['MODEL', 'TEXT', 'RAW', 'XML'])


@abstract
class PathElementA(EObject, metaclass=MetaEClass):

    def __init__(self, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

    def actualPathAbs(self):

        raise NotImplementedError('operation actualPathAbs(...) not yet implemented')

    def actualPath(self):

        raise NotImplementedError('operation actualPath(...) not yet implemented')

    def actualPathCwd(self):

        raise NotImplementedError('operation actualPathCwd(...) not yet implemented')


class Directory(PathElementA):

    name = EAttribute(eType=EString, derived=False, changeable=True, default_value='.')
    subdirectories = EReference(ordered=True, unique=True, containment=True, upper=-1)
    resources = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, subdirectories=None, resources=None, name=None, **kwargs):

        super().__init__(**kwargs)

        if name is not None:
            self.name = name

        if subdirectories:
            self.subdirectories.extend(subdirectories)

        if resources:
            self.resources.extend(resources)


@abstract
class FileResourceA(PathElementA):

    name = EAttribute(eType=EString, derived=False, changeable=True, default_value='newResource')
    isDirty = EAttribute(eType=EBoolean, derived=False, changeable=True)
    isLoaded = EAttribute(eType=EBoolean, derived=False, changeable=True, default_value=False)
    lastPersistentPath = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, *, name=None, isDirty=None, isLoaded=None, lastPersistentPath=None, **kwargs):

        super().__init__(**kwargs)

        if name is not None:
            self.name = name

        if isDirty is not None:
            self.isDirty = isDirty

        if isLoaded is not None:
            self.isLoaded = isLoaded

        if lastPersistentPath is not None:
            self.lastPersistentPath = lastPersistentPath


class Workspace(Directory):

    def __init__(self, **kwargs):

        super().__init__(**kwargs)


class ModelResource(FileResourceA):

    type = EAttribute(eType=FileResourceTypesE, derived=False,
                      changeable=True, default_value=FileResourceTypesE.MODEL)
    isMetaModel = EAttribute(eType=EBoolean, derived=False,
                             changeable=True, upper=-1, default_value=False)
    isWritable = EAttribute(eType=EBoolean, derived=False,
                            changeable=True, upper=-1, default_value=True)
    contents = EReference(ordered=True, unique=True, containment=True, upper=-1)

    def __init__(self, *, type=None, isMetaModel=None, isWritable=None, contents=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if isMetaModel:
            self.isMetaModel.extend(isMetaModel)

        if isWritable:
            self.isWritable.extend(isWritable)

        if contents:
            self.contents.extend(contents)


class TextResource(FileResourceA):

    type = EAttribute(eType=FileResourceTypesE, derived=False,
                      changeable=True, default_value=FileResourceTypesE.TEXT)
    lines = EAttribute(eType=EString, derived=False, changeable=True, upper=-1)

    def __init__(self, *, type=None, lines=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if lines:
            self.lines.extend(lines)


class RawResource(FileResourceA):

    type = EAttribute(eType=FileResourceTypesE, derived=False,
                      changeable=True, default_value=FileResourceTypesE.RAW)
    data = EAttribute(eType=EByte, derived=False, changeable=True, upper=-1)

    def __init__(self, *, type=None, data=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if data:
            self.data.extend(data)


class XmlResource(FileResourceA):

    type = EAttribute(eType=FileResourceTypesE, derived=False,
                      changeable=True, default_value=FileResourceTypesE.XML)
    lines = EAttribute(eType=EString, derived=False, changeable=True, upper=-1)
    document = EReference(ordered=True, unique=True, containment=True)

    def __init__(self, *, type=None, lines=None, document=None, **kwargs):

        super().__init__(**kwargs)

        if type is not None:
            self.type = type

        if lines:
            self.lines.extend(lines)

        if document is not None:
            self.document = document



'''
Example script demonstrating the use of the Essential Object Queries to read and modify ECORE models. 

Bjoern Annighoefer 2019
'''

#new imports
from ..query.query import Qry 
from ..command.command import Get, Set, Add, Rem, Mov, Clo, Crt, Crn, Sts, Chg, Gaa, Cal, Asc, Abc, Cmp, CloModes


#ERROR HANLDERS
class EoqCsvError(Exception):
    pass

class EoqCsvReadError(EoqCsvError):
    def __init__(self, errortype, message, eoqstring):
        self.errortype = errortype
        self.message = message
        self.eoqstring = eoqstring
        
           
#MAIN CLASS
class CsvConnector: 
    def __init__(self,domain):
        import csv
        self.elementIdLOT = {} #lookup table for elements ids
        self.elementLOT = {} #lookup table for elements
        self.idAttributeLOT = {} #lookup table id attributes
        self.dc = ';' #delimiter char
        self.qc = '"' #quote char
        self.sc = ',' #sub delimiter sign
        self.domain = domain
        self.quoting_state = csv.QUOTE_MINIMAL
        self.omit_empty_first_column = False
        self.csvWriter = None
        self.file = None

    # splits a string by the given delimiter. but respects quote characters (qc) 
    # for encapsulating strings containing the delimiter characters (dc)
    def _csvSplit(self,txt):
        import csv

        return list(csv.reader(txt, quotechar=self.qc, delimiter=self.dc,
                                    quoting=csv.QUOTE_ALL, skipinitialspace=True))

    def _tanslateArray(self,strArray,translationDictionary):
        n = len(strArray)
        for i in range(n):
            if strArray[i] in translationDictionary:
                strArray[i] = translationDictionary[strArray[i]]
        return strArray
    
    def _writeLine(self,lineArray):
        self.csvWriter.writerow(lineArray)

# obsolete as csvwriter handles this.
    #adds quote to a segment if it includes forbidden chars like a delimiter
#    def _csvOutputSegment(self,segmentTxt):
#
#        if(self.dc in segmentTxt or ' ' in segmentTxt):
#            return '%c%s%c'%(self.qc,segmentTxt,self.qc)
#        return segmentTxt #do nothing by default
    
    def getIdAttribute(self,element):
        idAttribute = None
        #clazz = element.eClass
        clazz = self.eoq.Retrieve(element,'(EObject:1)/eClass')
        if(clazz in self.idAttributeLOT):
            idAttribute = self.idAttributeLOT[clazz]
        else:
            idAttribute = self.eoq.Retrieve(clazz,'(EClass:1)/eAttributes{iD=true}')
            if(None == idAttribute):
                idAttribute = self.eoq.Retrieve(clazz,'(EClass:_)/eSuperTypes<-1>/eAttributes{iD=true}')
                if(len(idAttribute)>0):
                    idAttribute = idAttribute[0] #take only the first one. Multiple id attributes are not supported.
        self.idAttributeLOT[clazz] = idAttribute
        return idAttribute
          
    #return an element id
    def getElementId(self,element):
        if(element.v in self.elementIdLOT):
            return self.elementIdLOT[element.v]
        
        newElementId = None
        idAttribute = self.getIdAttribute(element);
        idAttributeName = self.eoq.Retrieve(idAttribute,'(EAttribute:1)/name')
        if idAttribute:
            idValue = self.eoq.Retrieve(element,'(EObject:1)/%s'%(idAttributeName))
            if idValue:
                if idValue in self.elementLOT:
                    existingElement = self.elementLOT[idValue]
                    print('WARNING: %s seems to have the non unique id %s. %s has the same id. Using the path instead.'%(element,idValue,existingElement))
                else:
                    newElementId = str(idValue)
                    self.elementLOT[newElementId] = element.v
        if newElementId is None:
            #no id available, so generate one by the path
            containerId = self.getContainerId(element)
            containedPath = self.getContainingPath(element,1)
            newElementId = "%s/%s"%(containerId,containedPath)
                
        self.elementIdLOT[element.v] = newElementId
        return newElementId
    
    def getContainerId(self,element):
        container = self.eoq.Retrieve(element,'(EObject:1)@CONTAINER')
        if container:
            return self.getElementId(container)
        return ''
    
    '''
    retrieves the path of an object
    depth is the number of anchestors to start from (use -1 to obtain the absolute path)
    '''
    def getContainingPath(self,obj,depth):
        path = self.eoq.Retrieve(obj,'(EObject:1)@CONTAININGFEATURENAME')
        index = self.eoq.Retrieve(obj,'(EObject:1)@INDEX')
        if index > 0:
            path += '[%d]'%(index)
                    
        depth = depth-1 #reduce depth if one level was resolved
        
        if depth: # depth is not 0
            container = self.eoq.Retrieve(obj,'(EObject:1)@CONTAINER')
            path = "%s/%s"%(self.getContainingPath(container,depth),path)
        
        return path

    def storeClass(self,package,clazz,root,ommitContainer,ommitContainingPath,attributeNameDictionary,classSectionBegin,classSectionEnd,ommitClassInfo):
        # clazzName = self.eoq.Retrieve(clazz,'(EClass:1)/name')
        clazzName = self.domain.Do(Get(Qry(clazz).Pth('name')))
        # allInstances = self.eoq.Retrieve(root,"(*:*)$%s"%(clazzName))
        # allInstances = self.domain.Do(Get(Qry(root).Sel(Qry().Equ(clazzName))))
        #allInstances = self.domain.Do(Get(Qry(root).Met('FEATUREVALUES').Sel(Qry(root).Met('FEATURENAMES').Equ(clazzName))))
        allInstances = self.domain.Do(Get(Qry(root).Cls(clazzName)))
        #print all attribute names
        if allInstances:
            #allAttributes = clazz.eAllAttributes()
            # ownAttributes = self.eoq.Retrieve(clazz,'(EClass:*)/eAttributes')
            ownAttributes = self.domain.Do(Get(Qry(clazz).Pth('eAttributes')))
            # superTypeAttributes = self.eoq.Retrieve(clazz,'(EClass:_)/eSuperTypes<-1>/eAttributes')
            #todo rekursiver supertype muss hier noch einmal überarbeitet werden
            superTypeAttributes = self.domain.Do(Get(Qry(clazz).Pth('eSuperTypes').Pth('eAttributes')))
            allAttributes = ownAttributes+superTypeAttributes
            #allReferences = clazz.eAllReferences()
            # todo hier muss der befehl noch einmal ausgearbeitet weden
            # ownNonContainingReferences = self.eoq.Retrieve(clazz,'(EClass:*)$EReference{containment=false}')
#            try:
            ownNonContainingReferences = self.domain.Do(Get(Qry(clazz).Cls('EReference').Sel(Qry().Pth('containment').Equ(False))))
#            except:
#                ownNonContainingReferences = []
            # superTypeNonContainingReferences = self.eoq.Retrieve(clazz,'(EClass:_)/eSuperTypes<-1>$EReference{containment=false}')
#            try:
            superTypeNonContainingReferences = self.domain.Do(Get(Qry(clazz).Pth('eSuperTypes').Cls('EReference').Sel(Qry().Pth('containment').Equ(False))))
#            except:
#                superTypeNonContainingReferences = []
            allNonContainingReferences = ownNonContainingReferences + superTypeNonContainingReferences

            # packageNsUri = self.eoq.Retrieve(package,'(EPackage:1)/nsURI')
            packageNsUri = self.domain.Do(Get(Qry(package).Pth('nsURI')))
            # className = self.eoq.Retrieve(clazz,'(EClass:1)/name')
            className = self.domain.Do(Get(Qry(clazz).Pth('name')))

            headerArray = []
            
            #print(classSectionBegin,end="",file=self.file)
            self._writeLine([classSectionBegin])
            if not ommitClassInfo:
                self._writeLine([packageNsUri,className])
            if not ommitContainer:
                headerArray.append("container")
                #print("container",end="",file=self.file)
            if not ommitContainingPath:
                headerArray.append("containedPath")
            #print("container%ccontainedPath"%(self.dc),end="",file=self.file)
            for attribute in allAttributes:
                # attributeName = self.eoq.Retrieve(attribute,'(EAttribute:1)/name')
                attributeName = self.domain.Do(Get(Qry(attribute).Pth('name')))
                headerArray.append(attributeName)
                #print("%c%s"%(self.dc,self._csvOutputSegment(attributeName)),end="",file=self.file)
            for reference in allNonContainingReferences:
                # referenceName = self.eoq.Retrieve(reference,'(EReference:1)/name')
                referenceName = self.domain.Do(Get(Qry(reference).Pth('name')))
                headerArray.append(referenceName)
                #print("%c%s"%(self.dc,self._csvOutputSegment(referenceName)),end="",file=self.file)
            #replace elements with names from the dictionary
            if attributeNameDictionary:
                headerArray = self._tanslateArray(headerArray,attributeNameDictionary)
            self._writeLine(headerArray)
            for instance in allInstances:
                lineArray = []
                if not ommitContainer:
                    lineArray.append(self.getContainerId(instance))
                if not ommitContainingPath:
#                     featureName = self.eoq.Retrieve(instance,'(EObject:1)@CONTAININGFEATURENAME')
#                     lineArray.append(featureName)
                    lineArray.append(self.getContainingPath(instance,1))
                #print('%s%c%s'%(self._csvOutputSegment(self.getContainerId(eoq,instance)),self.dc,self._csvOutputSegment(self.eoq.EoqPath(instance,1))),end="",file=self.file)
                for attribute in allAttributes:
                    # attributeName = self.eoq.Retrieve(attribute,'(EAttribute:1)/name')
                    attributeName = self.domain.Do(Get(Qry(attribute).Pth('name')))
                    # attributeValue = self.eoq.Retrieve(instance,'(%s:1)/%s'%(className,attributeName))
                    attributeValue = self.domain.Do(Get(Qry(instance).Pth(attributeName)))
                    attributeStr = str(attributeValue) if attributeValue else ''
                    lineArray.append(attributeStr)
                    #print('%c%s'%(self.dc,self._csvOutputSegment(attributeStr)),end="",file=self.file)
                for reference in allNonContainingReferences:
                    referenceName = self.eoq.Retrieve(reference,'(EReference:1)/name')
                    referedElements = self.eoq.Retrieve(instance,'(%s:*)/%s'%(className,referenceName))

                    elementIds = []
                    for element in referedElements:
                        elementIds.append(self.getElementId(element))

                if self.omit_empty_first_column and lineArray[0]: # send only lines that have a non empty first column
                    self._writeLine(lineArray)

            self._writeLine([classSectionEnd])

    def storePackage(self,package,root,ommitRootElement,ommitContainer,ommitContainingPath,attributeNameDictionary,
                     classSectionBegin,classSectionEnd,ommitClassInfo):
        #allClazzes = package.eClassifiers
        # allClazzes = self.eoq.Retrieve(package,'(EPackage:_)/eClassifiers$EClass{abstract=false}')
        allClasses = iter(self.domain.Do(Get(Qry(package).Pth('eClassifiers').Sel(Qry().Pth('abstract').Equ(False)))))
        if ommitRootElement:
            next(allClasses)
        for iClass in allClasses:
            self.storeClass(package,iClass,root,ommitContainer,ommitContainingPath,attributeNameDictionary,
                            classSectionBegin,classSectionEnd,ommitClassInfo)

            
        #look for subpackages
        # allSubpackages = self.eoq.Retrieve(package,'(EPackage:*)/eSubpackages')
        #allSubpackages = package.eSubpackages
        for subpackage in self.domain.Do(Get(Qry(package).Pth('eSubpackages'))):
            self.storePackage(subpackage,root,False,ommitContainer,ommitContainingPath,attributeNameDictionary,classSectionBegin,classSectionEnd,ommitClassInfo)
            
    def SaveAsCsv(self,modelRoot,filename,header='',ommitRootElement=False,ommitContainer=False,ommitContainingPath=False,
                  attributeNameDictionary={},classSectionBegin='',classSectionEnd='',ommitClassInfo=False):
       
        #reset members
        self.elementIdLOT = {}
        self.elementLOT = {}
        self.idAttributeLOT = {}
        
        # mainclass = self.eoq.Retrieve(modelRoot,'(*:1)/eClass')
        mainclass = self.domain.Do(Get(Qry(modelRoot).Pth('eClass')))
        mainclass = mainclass[0] if type(mainclass) == list else mainclass
        # mainpackage = self.eoq.Retrieve(mainclass,'(*:1)/ePackage')
        mainpackage = self.domain.Do(Get(Qry(mainclass).Pth('ePackage')))
        mainpackage = mainpackage[0] if type(mainpackage)==list else mainpackage
        
        self.startCsv(filename)
        for h in header:
            
            self._writeLine([h])
        self.storePackage(mainpackage,modelRoot,ommitRootElement,ommitContainer,ommitContainingPath,
                          attributeNameDictionary,classSectionBegin,classSectionEnd,ommitClassInfo)
        self.finish_csv()

    def startCsv(self, filename):
        import csv
        self.file = open(filename, 'w',newline='')
        # here we can add the correct delimiters
        self.csvWriter = csv.writer(self.file, delimiter=self.dc, lineterminator='\r\n',
                                    quotechar=self.qc, quoting=self.quoting_state)

    def finish_csv(self):
        self.file.close()



'''
Annighoefer 2019
'''

from .model import *
from .error import EoqValueError,EoqParseError

from pyecore.ecore import EObject

import types
from urllib.parse import quote,unquote  #for escaping strings

class ValueParser:
    @staticmethod
    def ValueToString(value):
        if(value.type in (ValueTypesE.INT,ValueTypesE.FLOAT)):
            return str(value.v)
        elif(value.type == ValueTypesE.BOOL):
            if(value.v):
                return 'true'
            else:
                return 'false'
        elif(value.type == ValueTypesE.STRING):
            return "'%s'"%(quote(value.v))
        elif(value.type == ValueTypesE.OBJECTREF):
            return '#%d'%(value.v)
        elif(value.type == ValueTypesE.HISTORYREF):
            return '$%d'%(value.v)
        elif(value.type == ValueTypesE.OBJECT):
            return '#%s'%(str(value.v))
        elif(value.type == ValueTypesE.EMPTY):
            return '%'
        elif(value.type == ValueTypesE.LIST):
            substrs = []
            for v in value.v:
                substrs += [ValueParser.ValueToString(v)]
            return '[%s]'%(','.join(substrs))
        else:
            raise EoqValueError(1,'Cannot convert EOQ value %s string'%(type(value)),value)
    
    @staticmethod
    def StringToValue(valueStr):
        #definition of the parser states
        VALUE_START = 1
        VALUE = 2
        VALUE_END = 3
        #start of parsing
        rootValueContainer = ListValue() #initialize the value with an empty list
        currentValueContainer = rootValueContainer
        state = VALUE_START
        valueStart = 0;
        valueEnd = 0;
        valueType = None
        nOpenBrakets = 0
        
        n = len(valueStr)
        for i in range(n):
            c = valueStr[i]
            if(VALUE_START == state):
                valueStart = i
                valueEnd = i
                if(('0' <= c and c <= '9') or '-' == c):
                    valueType = ValueTypesE.INT
                    state = VALUE
                elif('.'==c):
                    valueType = ValueTypesE.FLOAT
                    state = VALUE
                elif('t'==c or 'f'==c):
                    valueType = ValueTypesE.BOOL
                    state = VALUE
                elif('#'==c):
                    valueStart = i+1
                    valueEnd = i+1
                    valueType = ValueTypesE.OBJECTREF
                    state = VALUE
                elif('\''==c):
                    valueStart = i+1
                    valueEnd = i+1
                    valueType = ValueTypesE.STRING
                    state = VALUE
                elif('%'==c):
                    valueStart = i+1
                    valueEnd = i+1
                    valueType = ValueTypesE.EMPTY
                    state = VALUE
                elif('$'==c):
                    valueStart = i+1
                    valueEnd = i+1
                    valueType = ValueTypesE.HISTORYREF
                    state = VALUE
                elif('['==c):
                    valueType = ValueTypesE.LIST
                    nOpenBrakets += 1
                    sublist = ListValue()
                    currentValueContainer.v.add(sublist)
                    currentValueContainer = sublist
                    state = VALUE_START
                elif(']'==c):
                    #nothing needs to be added because it has already been added
                    nOpenBrakets -= 1
                    currentValueContainer = currentValueContainer.eContainer()
                    state = VALUE_END
                else:
                    raise EoqParseError(10,'Unexpected character \'%c\' at %d'%(c,i),valueStr)
            elif(VALUE == state):
                if(']'==c):
                    currentValueContainer.v.add(ValueParser.__StringToPritiveValue(valueStr[valueStart:valueEnd+1],valueType))
                    nOpenBrakets -= 1
                    currentValueContainer = currentValueContainer.eContainer()
                    state = VALUE_END
                elif(','==c and nOpenBrakets>0):
                    #TODO evaluate the value
                    currentValueContainer.v.add(ValueParser.__StringToPritiveValue(valueStr[valueStart:valueEnd+1],valueType))
                    state = VALUE_START
                elif((ValueTypesE.INT==valueType or ValueTypesE.FLOAT==valueType) and ('0' <= c and c <= '9')):
                    valueEnd = i
                elif(ValueTypesE.OBJECTREF==valueType and ('0' <= c and c <= '9')):
                    valueEnd = i
                elif(ValueTypesE.HISTORYREF==valueType and (('0' <= c and c <= '9') or c == '-')):
                    valueEnd = i
                elif(ValueTypesE.INT==valueType and'.'==c):
                    valueEnd = i
                    valueType = ValueTypesE.FLOAT
                elif(ValueTypesE.BOOL==valueType and c.isalpha()):
                    valueEnd = i
                elif(ValueTypesE.STRING==valueType and '\''!=c):
                    valueEnd = i
                elif(ValueTypesE.STRING==valueType and '\''==c):
                    valueEnd = i-1 #necessary for empty string
                else:
                    raise EoqParseError(11,'Unexpected character \'%c\' at %d'%(c,i),valueStr)
            elif(VALUE_END==state):
                if(']'==c):
                    nOpenBrakets -= 1
                    currentValueContainer = currentValueContainer.eContainer()
                elif(','==c and nOpenBrakets>0):
                    state = VALUE_START
                else:
                    raise EoqParseError(12,'Unexpected character \'%c\' at %d'%(c,i),valueStr)
        
        if(nOpenBrakets>0):
            raise EoqParseError(13,'%d lists are not closed'%(nOpenBrakets),valueStr)
        
        #process the last value
        if(VALUE==state):
            currentValueContainer.v.add(ValueParser.__StringToPritiveValue(valueStr[valueStart:valueEnd+1],valueType))        
        
        return rootValueContainer.v[0]
    
    @staticmethod
    def __StringToPritiveValue(valueStr,valueType):
        value = None
        if(valueType == ValueTypesE.INT):
            value = IntValue()
            value.v = int(valueStr)
        elif(valueType == ValueTypesE.FLOAT):
            value = FloatValue()
            value.v = float(valueStr)
        elif(valueType == ValueTypesE.STRING):
            value = StringValue()
            value.v = unquote(valueStr)
        elif(valueType == ValueTypesE.BOOL):
            value = BoolValue()
            if('true' == valueStr):
                value.v = True
            elif('false' == valueStr):
                value.v = False
            else:
                raise EoqParseError(20,"Unsupported boolean value '%s'. Should be 'true' or 'false'."%(valueStr),valueStr)
        elif(valueType == ValueTypesE.OBJECTREF):
            value = ObjectRefValue()
            value.v = int(valueStr)
        elif(valueType == ValueTypesE.EMPTY):
            value = EmptyValue()
        elif(valueType == ValueTypesE.HISTORYREF):
            value = HistoryRefValue()
            value.v = int(valueStr)
        else:
            raise EoqParseError(21,"Unknown value type '%s'."%(ValueTypesE.to_string(valueType)),valueStr)
        return value
    
    @staticmethod
    def PythonToValue(pvalue):
        if(isinstance(pvalue,bool)): #int is instance of bool, therefore bool must be first.
            return BoolValue(v=pvalue)
        elif(isinstance(pvalue,int)):
            return IntValue(v=pvalue)
        elif(isinstance(pvalue,float)):
            return FloatValue(v=pvalue)
        elif(isinstance(pvalue,str)):
            return StringValue(v=pvalue)
        elif(isinstance(pvalue,ObjectRefValue)):
            return pvalue
        elif(isinstance(pvalue,HistoryRefValue)):
            return pvalue
        elif(isinstance(pvalue,types.FunctionType) or isinstance(pvalue,types.MethodType)):
            return OperationValue(v=pvalue)
        elif(isinstance(pvalue,EObject)):
            return ObjectValue(v=pvalue)
        elif(None == pvalue):
            return EmptyValue()
        elif(isinstance(pvalue,list)):
            value = ListValue()
            for v in pvalue:
                value.v.add(ValueParser.PythonToValue(v))
            return value
        else:
            raise EoqValueError(31,'Cannot convert python data type %s to EOQ value'%(type(pvalue).__name__),pvalue)
        
    @staticmethod
    def ValueToPython(value):
        if(value.type in (ValueTypesE.INT,ValueTypesE.FLOAT,ValueTypesE.BOOL,ValueTypesE.STRING,ValueTypesE.OBJECT)):
            return value.v
        elif(value.type in (ValueTypesE.OBJECTREF,ValueTypesE.HISTORYREF,ValueTypesE.OPERATION)):
            return value
        elif(value.type == ValueTypesE.OPERATION):
            return value.v
        elif(value.type == ValueTypesE.EMPTY):
            return None
        elif(value.type == ValueTypesE.LIST):
            pvalue = []
            for v in value.v:
                pvalue += [ValueParser.ValueToPython(v)]
            return pvalue
        else:
            raise EoqValueError(32,'Cannot convert EOQ value %s to python data type'%(type(value).__name__),value)
        
    @staticmethod
    def CloneValue(value):
        clonedValue = None
        if(value.type == ValueTypesE.LIST):
            clonedValue = ListValue()
            for v in value.v:
                clonedValue.v.add(ValueParser.CloneValue(v))
        elif(value.type == ValueTypesE.OBJECTREF):
            clonedValue = ObjectRefValue(v=value.v)
        elif(value.type == ValueTypesE.INT):
            clonedValue = IntValue(v=value.v)
        elif(value.type == ValueTypesE.STRING):
            clonedValue = StringValue(v=value.v)
        elif(value.type == ValueTypesE.FLOAT):
            clonedValue = FloatValue(v=value.v)
        elif(value.type == ValueTypesE.BOOL):
            clonedValue = BoolValue(v=value.v)
        elif(value.type == ValueTypesE.EMPTY):
            clonedValue = EmptyValue()
        elif(value.type == ValueTypesE.HISTORYREF):
            clonedValue = HistoryRefValue(v=value.v)
        elif(value.type == ValueTypesE.OPERATION):
            clonedValue = OperationValue(v=value.v)
        elif(value.type == ValueTypesE.OBJECT):
            clonedValue == ObjectValue(v=value.v)
        return clonedValue
        
                